"""Type flo.params.brickstorageheater, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.enums import RecognizedCurrencyUnit
from gwatn.enums import RecognizedTemperatureUnit


class RecognizedCurrencyUnit000SchemaEnum:
    enum_name: str = "recognized.currency.unit.000"
    symbols: List[str] = [
        "00000000",
        "e57c5143",
        "f7b38fc5",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class RecognizedCurrencyUnit000(StrEnum):
    UNKNOWN = auto()
    USD = auto()
    GBP = auto()

    @classmethod
    def default(cls) -> "RecognizedCurrencyUnit000":
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class RecognizedCurrencyUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> RecognizedCurrencyUnit:
        if not RecognizedCurrencyUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to RecognizedCurrencyUnit000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )

    @classmethod
    def local_to_type(cls, recognized_currency_unit: RecognizedCurrencyUnit) -> str:
        if not isinstance(recognized_currency_unit, RecognizedCurrencyUnit):
            raise SchemaError(
                f"{recognized_currency_unit} must be of type {RecognizedCurrencyUnit}"
            )
        versioned_enum = as_enum(
            recognized_currency_unit,
            RecognizedCurrencyUnit000,
            RecognizedCurrencyUnit000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, RecognizedCurrencyUnit000] = {
        "00000000": RecognizedCurrencyUnit000.UNKNOWN,
        "e57c5143": RecognizedCurrencyUnit000.USD,
        "f7b38fc5": RecognizedCurrencyUnit000.GBP,
    }

    versioned_enum_to_type_dict: Dict[RecognizedCurrencyUnit000, str] = {
        RecognizedCurrencyUnit000.UNKNOWN: "00000000",
        RecognizedCurrencyUnit000.USD: "e57c5143",
        RecognizedCurrencyUnit000.GBP: "f7b38fc5",
    }


class RecognizedTemperatureUnit000SchemaEnum:
    enum_name: str = "recognized.temperature.unit.000"
    symbols: List[str] = [
        "00000000",
        "6f16ee63",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class RecognizedTemperatureUnit000(StrEnum):
    C = auto()
    F = auto()

    @classmethod
    def default(cls) -> "RecognizedTemperatureUnit000":
        return cls.C

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class RecognizedTemperatureUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> RecognizedTemperatureUnit:
        if not RecognizedTemperatureUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to RecognizedTemperatureUnit000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum,
            RecognizedTemperatureUnit,
            RecognizedTemperatureUnit.default(),
        )

    @classmethod
    def local_to_type(
        cls, recognized_temperature_unit: RecognizedTemperatureUnit
    ) -> str:
        if not isinstance(recognized_temperature_unit, RecognizedTemperatureUnit):
            raise SchemaError(
                f"{recognized_temperature_unit} must be of type {RecognizedTemperatureUnit}"
            )
        versioned_enum = as_enum(
            recognized_temperature_unit,
            RecognizedTemperatureUnit000,
            RecognizedTemperatureUnit000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, RecognizedTemperatureUnit000] = {
        "00000000": RecognizedTemperatureUnit000.C,
        "6f16ee63": RecognizedTemperatureUnit000.F,
    }

    versioned_enum_to_type_dict: Dict[RecognizedTemperatureUnit000, str] = {
        RecognizedTemperatureUnit000.C: "00000000",
        RecognizedTemperatureUnit000.F: "6f16ee63",
    }


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


class FloParamsBrickstorageheater(BaseModel):
    """.

    FloParams for the BrickStorageHeater AtomicTNode strategy.
    [More info](https://gridworks-atn.readthedocs.io/en/latest/brick-storage-heater.html).
    """

    MaxBrickTempC: int = Field(
        title="MaxBrickTempC",
        default=190,
    )
    RatedMaxPowerKw: float = Field(
        title="RatedMaxPowerKw",
        default=13.5,
    )
    ROff: float = Field(
        title="ROff",
        default=0.08,
    )
    ROn: float = Field(
        title="ROn",
        default=0.15,
    )
    RoomTempF: int = Field(
        title="RoomTempF",
        default=70,
    )
    CurrencyUnit: RecognizedCurrencyUnit = Field(
        title="CurrencyUnit",
        default=RecognizedCurrencyUnit.USD,
    )
    TempUnit: RecognizedTemperatureUnit = Field(
        title="TempUnit",
        default=RecognizedTemperatureUnit.F,
    )
    TimezoneString: str = Field(
        title="TimezoneString",
        default="US/Eastern",
    )
    HomeCity: str = Field(
        title="HomeCity",
        default="MILLINOCKET_ME",
    )
    IsRegulating: bool = Field(
        title="IsRegulating",
        default=False,
    )
    StorageSteps: int = Field(
        title="StorageSteps",
        default=100,
    )
    SliceDurationMinutes: List[int] = Field(
        title="SliceDurationMinutes",
        default=[60],
    )
    PowerRequiredByHouseFromSystemAvgKwList: List[float] = Field(
        title="PowerRequiredByHouseFromSystemAvgKwList",
        default=[3.42],
    )
    C: Optional[float] = Field(
        title="C",
        default=200,
    )
    RealtimeElectricityPrice: List[float] = Field(
        title="RealtimeElectricityPrice",
        default=[10.35],
    )
    OutsideTempF: List[float] = Field(
        title="OutsideTempF",
        default=[-5.1],
    )
    DistributionPrice: List[float] = Field(
        title="DistributionPrice",
        default=[40.0],
    )
    RtElecPriceUid: str = Field(
        title="RtElecPriceUid",
    )
    RegulationPrice: List[float] = Field(
        title="RegulationPrice",
        default=[25.3],
    )
    WeatherUid: str = Field(
        title="WeatherUid",
    )
    DistPriceUid: Optional[str] = Field(
        title="DistPriceUid",
        default=None,
    )
    RegPriceUid: Optional[str] = Field(
        title="RegPriceUid",
        default=None,
    )
    StartYearUtc: int = Field(
        title="StartYearUtc",
        default=2020,
    )
    StartMonthUtc: int = Field(
        title="StartMonthUtc",
        default=1,
    )
    StartDayUtc: int = Field(
        title="StartDayUtc",
        default=1,
    )
    StartHourUtc: int = Field(
        title="StartHourUtc",
        default=0,
    )
    StartMinuteUtc: int = Field(
        title="StartMinuteUtc",
        default=0,
    )
    StartingStoreIdx: int = Field(
        title="StartingStoreIdx",
        default=50,
    )
    AmbientPowerInKw: float = Field(
        title="AmbientPowerInKw",
        default=1.25,
    )
    HouseWorstCaseTempF: int = Field(
        title="HouseWorstCaseTempF",
        default=-7,
    )
    GNodeAlias: str = Field(
        title="GNodeAlias",
    )
    FloParamsUid: str = Field(
        title="FloParamsUid",
    )
    TypeName: Literal["flo.params.brickstorageheater"] = "flo.params.brickstorageheater"
    Version: str = "000"

    @validator("CurrencyUnit")
    def _check_currency_unit(cls, v: RecognizedCurrencyUnit) -> RecognizedCurrencyUnit:
        return as_enum(v, RecognizedCurrencyUnit, RecognizedCurrencyUnit.UNKNOWN)

    @validator("TempUnit")
    def _check_temp_unit(
        cls, v: RecognizedTemperatureUnit
    ) -> RecognizedTemperatureUnit:
        return as_enum(v, RecognizedTemperatureUnit, RecognizedTemperatureUnit.C)

    @validator("RtElecPriceUid")
    def _check_rt_elec_price_uid(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"RtElecPriceUid failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("WeatherUid")
    def _check_weather_uid(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"WeatherUid failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("DistPriceUid")
    def _check_dist_price_uid(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"DistPriceUid failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("RegPriceUid")
    def _check_reg_price_uid(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"RegPriceUid failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("GNodeAlias")
    def _check_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"GNodeAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("FloParamsUid")
    def _check_flo_params_uid(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"FloParamsUid failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["CurrencyUnit"]
        CurrencyUnit = as_enum(
            self.CurrencyUnit, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )
        d["CurrencyUnitGtEnumSymbol"] = RecognizedCurrencyUnitMap.local_to_type(
            CurrencyUnit
        )
        del d["TempUnit"]
        TempUnit = as_enum(
            self.TempUnit,
            RecognizedTemperatureUnit,
            RecognizedTemperatureUnit.default(),
        )
        d["TempUnitGtEnumSymbol"] = RecognizedTemperatureUnitMap.local_to_type(TempUnit)
        if d["C"] is None:
            del d["C"]
        if d["DistPriceUid"] is None:
            del d["DistPriceUid"]
        if d["RegPriceUid"] is None:
            del d["RegPriceUid"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class FloParamsBrickstorageheater_Maker:
    type_name = "flo.params.brickstorageheater"
    version = "000"

    def __init__(
        self,
        max_brick_temp_c: int,
        rated_max_power_kw: float,
        r_off: float,
        r_on: float,
        room_temp_f: int,
        currency_unit: RecognizedCurrencyUnit,
        temp_unit: RecognizedTemperatureUnit,
        timezone_string: str,
        home_city: str,
        is_regulating: bool,
        storage_steps: int,
        slice_duration_minutes: List[int],
        power_required_by_house_from_system_avg_kw_list: List[float],
        c: Optional[float],
        realtime_electricity_price: List[float],
        outside_temp_f: List[float],
        distribution_price: List[float],
        rt_elec_price_uid: str,
        regulation_price: List[float],
        weather_uid: str,
        dist_price_uid: Optional[str],
        reg_price_uid: Optional[str],
        start_year_utc: int,
        start_month_utc: int,
        start_day_utc: int,
        start_hour_utc: int,
        start_minute_utc: int,
        starting_store_idx: int,
        ambient_power_in_kw: float,
        house_worst_case_temp_f: int,
        g_node_alias: str,
        flo_params_uid: str,
    ):
        self.tuple = FloParamsBrickstorageheater(
            MaxBrickTempC=max_brick_temp_c,
            RatedMaxPowerKw=rated_max_power_kw,
            ROff=r_off,
            ROn=r_on,
            RoomTempF=room_temp_f,
            CurrencyUnit=currency_unit,
            TempUnit=temp_unit,
            TimezoneString=timezone_string,
            HomeCity=home_city,
            IsRegulating=is_regulating,
            StorageSteps=storage_steps,
            SliceDurationMinutes=slice_duration_minutes,
            PowerRequiredByHouseFromSystemAvgKwList=power_required_by_house_from_system_avg_kw_list,
            C=c,
            RealtimeElectricityPrice=realtime_electricity_price,
            OutsideTempF=outside_temp_f,
            DistributionPrice=distribution_price,
            RtElecPriceUid=rt_elec_price_uid,
            RegulationPrice=regulation_price,
            WeatherUid=weather_uid,
            DistPriceUid=dist_price_uid,
            RegPriceUid=reg_price_uid,
            StartYearUtc=start_year_utc,
            StartMonthUtc=start_month_utc,
            StartDayUtc=start_day_utc,
            StartHourUtc=start_hour_utc,
            StartMinuteUtc=start_minute_utc,
            StartingStoreIdx=starting_store_idx,
            AmbientPowerInKw=ambient_power_in_kw,
            HouseWorstCaseTempF=house_worst_case_temp_f,
            GNodeAlias=g_node_alias,
            FloParamsUid=flo_params_uid,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: FloParamsBrickstorageheater) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> FloParamsBrickstorageheater:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> FloParamsBrickstorageheater:
        d2 = dict(d)
        if "MaxBrickTempC" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MaxBrickTempC")
        if "RatedMaxPowerKw" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RatedMaxPowerKw")
        if "ROff" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ROff")
        if "ROn" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ROn")
        if "RoomTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RoomTempF")
        if "CurrencyUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing CurrencyUnitGtEnumSymbol")
        if (
            d2["CurrencyUnitGtEnumSymbol"]
            in RecognizedCurrencyUnit000SchemaEnum.symbols
        ):
            d2["CurrencyUnit"] = RecognizedCurrencyUnitMap.type_to_local(
                d2["CurrencyUnitGtEnumSymbol"]
            )
        else:
            d2["CurrencyUnit"] = RecognizedCurrencyUnit.default()
        if "TempUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TempUnitGtEnumSymbol")
        if d2["TempUnitGtEnumSymbol"] in RecognizedTemperatureUnit000SchemaEnum.symbols:
            d2["TempUnit"] = RecognizedTemperatureUnitMap.type_to_local(
                d2["TempUnitGtEnumSymbol"]
            )
        else:
            d2["TempUnit"] = RecognizedTemperatureUnit.default()
        if "TimezoneString" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TimezoneString")
        if "HomeCity" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HomeCity")
        if "IsRegulating" not in d2.keys():
            raise SchemaError(f"dict {d2} missing IsRegulating")
        if "StorageSteps" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StorageSteps")
        if "SliceDurationMinutes" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SliceDurationMinutes")
        if "PowerRequiredByHouseFromSystemAvgKwList" not in d2.keys():
            raise SchemaError(
                f"dict {d2} missing PowerRequiredByHouseFromSystemAvgKwList"
            )
        if "C" not in d2.keys():
            d2["C"] = None
        if "RealtimeElectricityPrice" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RealtimeElectricityPrice")
        if "OutsideTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing OutsideTempF")
        if "DistributionPrice" not in d2.keys():
            raise SchemaError(f"dict {d2} missing DistributionPrice")
        if "RtElecPriceUid" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RtElecPriceUid")
        if "RegulationPrice" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RegulationPrice")
        if "WeatherUid" not in d2.keys():
            raise SchemaError(f"dict {d2} missing WeatherUid")
        if "DistPriceUid" not in d2.keys():
            d2["DistPriceUid"] = None
        if "RegPriceUid" not in d2.keys():
            d2["RegPriceUid"] = None
        if "StartYearUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartYearUtc")
        if "StartMonthUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartMonthUtc")
        if "StartDayUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartDayUtc")
        if "StartHourUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartHourUtc")
        if "StartMinuteUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartMinuteUtc")
        if "StartingStoreIdx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartingStoreIdx")
        if "AmbientPowerInKw" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AmbientPowerInKw")
        if "HouseWorstCaseTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HouseWorstCaseTempF")
        if "GNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeAlias")
        if "FloParamsUid" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FloParamsUid")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return FloParamsBrickstorageheater(
            MaxBrickTempC=d2["MaxBrickTempC"],
            RatedMaxPowerKw=d2["RatedMaxPowerKw"],
            ROff=d2["ROff"],
            ROn=d2["ROn"],
            RoomTempF=d2["RoomTempF"],
            CurrencyUnit=d2["CurrencyUnit"],
            TempUnit=d2["TempUnit"],
            TimezoneString=d2["TimezoneString"],
            HomeCity=d2["HomeCity"],
            IsRegulating=d2["IsRegulating"],
            StorageSteps=d2["StorageSteps"],
            SliceDurationMinutes=d2["SliceDurationMinutes"],
            PowerRequiredByHouseFromSystemAvgKwList=d2[
                "PowerRequiredByHouseFromSystemAvgKwList"
            ],
            C=d2["C"],
            RealtimeElectricityPrice=d2["RealtimeElectricityPrice"],
            OutsideTempF=d2["OutsideTempF"],
            DistributionPrice=d2["DistributionPrice"],
            RtElecPriceUid=d2["RtElecPriceUid"],
            RegulationPrice=d2["RegulationPrice"],
            WeatherUid=d2["WeatherUid"],
            DistPriceUid=d2["DistPriceUid"],
            RegPriceUid=d2["RegPriceUid"],
            StartYearUtc=d2["StartYearUtc"],
            StartMonthUtc=d2["StartMonthUtc"],
            StartDayUtc=d2["StartDayUtc"],
            StartHourUtc=d2["StartHourUtc"],
            StartMinuteUtc=d2["StartMinuteUtc"],
            StartingStoreIdx=d2["StartingStoreIdx"],
            AmbientPowerInKw=d2["AmbientPowerInKw"],
            HouseWorstCaseTempF=d2["HouseWorstCaseTempF"],
            GNodeAlias=d2["GNodeAlias"],
            FloParamsUid=d2["FloParamsUid"],
            TypeName=d2["TypeName"],
            Version="000",
        )



""" BrickStorageHeater AtomicTNode Strategy.

 This is a heating and bidding strategy for a thermal storage heater that is
 designed to heat part or all of a single room. The storage medium is ceramic
 bricks and the heating source is resistive elements embedded in those bricks.

 This kind of heater is often called a Night Storage Heater in the UK, and
 is often also referred to (ambigiously, since there are other kinds) as
 an Electric Thermal Storage (ETS) heater.
 """
import functools
import logging
import math
import random
import time
import uuid
from typing import no_type_check

import dotenv
import gridworks.algo_utils as algo_utils
import pendulum
import requests
import satn.strategies.heatpumpwithbooststore.atn_utils as atn_utils
import satn.strategies.heatpumpwithbooststore.dev_io as dev_io
from algosdk import encoding
from algosdk.future import transaction
from algosdk.v2client.algod import AlgodClient
from gridworks.algo_utils import BasicAccount
from gridworks.utils import RestfulResponse
from satn.strategies.heatpumpwithbooststore.atn_utils import SlotStuff
from satn.strategies.heatpumpwithbooststore.edge import (
    Edge__SpaceHeat__Water_HeatPumpAndBoost__Alpha as Edge,
)
from satn.strategies.heatpumpwithbooststore.flo import (
    HeatPumpWithBoostStore__Flo as Flo,
)

import gwatn.config as config
from gwatn.atn_actor_base import AtnActorBase
from gwatn.data_classes.market_type import Rt60Gate30B
from gwatn.enums import GNodeRole
from gwatn.enums import MessageCategory
from gwatn.enums import MessageCategorySymbol
from gwatn.enums import UniverseType
from gwatn.types import AcceptedBid_Maker
from gwatn.types import AtnParamsBrickstorageheater as AtnParams
from gwatn.types import AtnParamsReport_Maker
from gwatn.types import HeartbeatA
from gwatn.types import HeartbeatA_Maker
from gwatn.types import LatestPrice
from gwatn.types import MarketSlot
from gwatn.types import MarketTypeGt
from gwatn.types import MarketTypeGt_Maker
from gwatn.types import PriceQuantityUnitless
from gwatn.types import Ready_Maker
from gwatn.types import SimplesimSnapshotBrickstorageheater_Maker as Snapshot_Maker
from gwatn.types import SimTimestep


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.WARNING)


class Atn__BrickStorageHeater(AtnActorBase):
    """AtomicTNode HeatPumpWithBoostStore strategy for thermal storage heat pump
    space heating system"""

    def __init__(
        self,
        settings: config.AtnSettings = config.AtnSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        super().__init__(settings)
        LOGGER.info("Initializing HeatPumpWithBoostStore Atn")
        self.algo_acct: BasicAccount = BasicAccount(settings.sk.get_secret_value())
        self.algo_client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        self.atn_params: AtnParams = atn_utils.dummy_atn_params()
        self.market_type: MarketTypeGt = MarketTypeGt_Maker.dc_to_tuple(Rt60Gate30B)
        self.active_run: SlotStuff = atn_utils.dummy_slot_stuff(
            slot=self.active_slot(self.market_type)
        )
        self.next_run: SlotStuff = atn_utils.dummy_slot_stuff(
            slot=self.next_slot(self.market_type)
        )
        self.store_kwh: float = 0
        self.heatpump_kw: float = 0
        self.boost_kw: float = 0
        self.cop: float = 2.5
        self.store_update_time: float = self.time()
        self.latest_price_received_time: float = 0
        self.latest_price_dollars_per_mwh: float = 10**6
        self._first_start_finished: bool = False

    @no_type_check
    def strategy_rabbit_startup(self) -> None:
        mm_alias_lrh = self.settings.market_maker_alias.replace(".", "-")
        rjb = MessageCategorySymbol.rjb.value
        binding = f"{rjb}.{mm_alias_lrh}.marketmaker.latest-price.{self.market_type.Name.value}"
        cb = functools.partial(self.on_marketprice_bindok, binding=binding)
        self._consume_channel.queue_bind(
            self.queue_name, "marketmakermic_tx", routing_key=binding, callback=cb
        )

        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex="0"
        ).tuple
        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )
        d = pendulum.from_timestamp(time.time())
        LOGGER.warning(
            f"[{self.short_alias}] Sent first heartbeat to super: {d.minute}:{d.second}.{d.microsecond}"
        )

    @no_type_check
    def on_marketprice_bindok(self, _unused_frame, binding) -> None:
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")

    def heartbeat_received(self, from_alias: str, ping: HeartbeatA):
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex=ping.MyHex
        ).tuple

        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )

        LOGGER.debug(
            f"[{self.short_alias}] Sent HB: SuHex {pong.YourLastHex}, AtnHex {pong.MyHex}"
        )

    def new_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("----------------------------------------------------")
        # LOGGER.info(f"[{self.time_str()}: {self.short_alias}] NEW TIMESTEP")
        # LOGGER.info("----------------------------------------------------")

        # This gets called on the first timestep
        if atn_utils.is_dummy_atn_params(self.atn_params):
            self.get_initial_params()
            LOGGER.info("Correcting runs on initial timestep")
            self.active_run = atn_utils.dummy_slot_stuff(
                slot=self.last_slot(self.market_type),
            )
            try:
                self.next_run = dev_io.initialize_slot_stuff(
                    slot=self.active_slot(self.market_type),
                    atn_params=self.atn_params,
                )
            except Exception as e:
                LOGGER.warning(
                    f"Failed to initialize FloParams! Not creating a next_run {e}"
                )
                return

        try:
            self.update_store_level()
        except Exception:
            LOGGER.exception("Error in update_store_level")

        # Note: compressing the bid and the market start into the same time,
        # since time steps are only once per slot right now
        if self.time() >= self.next_run.Slot.StartUnixS:
            self.active_run = self.next_run
            self.active_run.BidParams.StartingStoreIdx = self.store_idx()
            try:
                self.get_bid()
            except Exception:
                LOGGER.exception("Error in get_bid")
            self.next_run = dev_io.initialize_slot_stuff(
                slot=self.next_slot(self.market_type),
                atn_params=self.atn_params,
            )
            try:
                self.submit_bid(self.active_run)
            except Exception:
                LOGGER.exception("Error in submit_bid")
        if self.active_run.Price is not None:
            self.respond_to_price()

    def repeat_timestep(self, payload: SimTimestep) -> None:
        LOGGER.info(f"[{self.time_str()}: {self.short_alias}] Timestep received again")

    def latest_price_from_market_maker(self, payload: LatestPrice) -> None:
        self.mm_payload = payload

        if payload.FromGNodeAlias != self.settings.market_maker_alias:
            LOGGER.warning(
                f"Received price from {payload.FromGNodeAlias}."
                f" Expected {self.settings.market_maker_alias} "
            )
            return
        slot = atn_utils.market_slot_from_name(payload.MarketSlotName)

        # time_str = pendulum.from_timestamp(slot.StartUnixS).strftime("%m/%d/%Y, %H:%M")
        # LOGGER.info("----------------------------------------------------")
        # LOGGER.info(
        #     f"[{self.time_str()}: {self.short_alias}] LATEST PRICE From MarketMaker: "
        #     f"${payload.PriceTimes1000 / 1000}/Mwh for slot starting {time_str}"
        # )
        # LOGGER.info("----------------------------------------------------")
        if slot.Type != self.market_type:
            LOGGER.warning(
                f"Received price for {slot.Type}, only participate in {self.market_type}"
            )
            return

        if slot == self.active_run.Slot:
            # LOGGER.info("Received price from MarketMaker. Adding price to active run")
            self.active_run.Price = payload.PriceTimes1000 / 1000
        elif slot == self.next_run.Slot:
            # this happens when the price arrives before the timestep
            # LOGGER.info("Received price from MarketMaker. Adding price to next run")
            self.next_run.Price = payload.PriceTimes1000 / 1000
        else:
            raise Exception("HUH. did not add price to a run")

        if self.active_run.Price is not None:
            self.respond_to_price()

    def respond_to_price(self) -> None:
        try:
            self.update_power_levels()
        except Exception:
            LOGGER.exception("Error in update_power_levels")
        payload = Snapshot_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            boost_power_kw_times1000=int(self.boost_kw * 1000),
            heatpump_power_kw_times1000=int(self.heatpump_kw * 1000),
            cop_times10=int(self.cop * 10),
            store_kwh=int(self.store_kwh),
            max_store_kwh=int(atn_utils.get_max_store_kwh_th(self.atn_params)),
            about_terminal_asset_alias=self.alias + ".ta",
        ).tuple
        # pprint(payload)
        self.send_message(
            payload=payload, message_category=MessageCategory.RabbitJsonBroadcast
        )

        payload = Ready_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            time_unix_s=int(self.time()),
        ).tuple
        self.send_message(
            payload=payload,
            to_role=GNodeRole.TimeCoordinator,
            to_g_node_alias=self.settings.time_coordinator_alias,
        )

    #########################
    # Terminal Asset State
    #########################

    def update_power_levels(self) -> None:
        # use price and Flo results to figure out heatpump and boost levels
        if self.active_run.Flo is None:
            return
        node = self.active_run.Flo.node[0][self.active_run.Flo.starting_store_idx]
        edge: Edge = list(
            filter(
                lambda x: x.end_idx == node.best_next_idx,
                self.active_run.Flo.edges[node],
            )
        )[0]
        self.heatpump_kw = edge.hp_electricity_avg_kw
        self.boost_kw = edge.boost_electricity_used_avg_kw
        self.cop = self.active_run.Flo.cop[0]
        # LOGGER.info(
        #     f"[{self.time_str()}: {self.short_alias}] Updating state: heatpump {round(self.heatpump_kw,2)}"
        # )

    def update_store_level(self) -> None:
        if atn_utils.is_dummy_slot_stuff(self.active_run):
            LOGGER.info(f"Not updating storage - active run is a dummy")
            self.store_update_time = self.time()
            return
        duration_minutes = int((self.time() - self.store_update_time) / 60)
        if duration_minutes != self.market_type.DurationMinutes:
            raise Exception(
                f"In update_store_level: Expected {self.market_type.DurationMinutes} minutes, got {duration_minutes} minutes."
            )
        flo = self.active_run.Flo
        if flo is None:
            raise Exception(
                f"In update_store_level: Expected Flo for last_run {self.active_run}!"
            )
        # TODO: find next step in graph based on the Flo, the Bid, and the actual price
        if flo.starting_store_idx != self.store_idx():
            raise Exception(
                f"flo.starting_idx is {flo.starting_store_idx} but self.store_idx is {self.store_idx()}!"
            )
        try:
            store_idx = flo.node[0][flo.starting_store_idx].best_next_idx
        except:
            raise Exception(f"Flo node did not have best_node_idx!")
        # self.store_kwh = ...
        self.store_kwh = (
            store_idx
            * atn_utils.get_max_store_kwh_th(self.atn_params)
            / self.atn_params.StorageSteps
        )
        # LOGGER.info(
        #     f"[{self.time_str()}: {self.short_alias}] Storage level {round(self.store_kwh,2)} kWh, "
        #     f"StoreIdx {self.active_run.BidParams.StartingStoreIdx}"
        # )
        self.store_update_time = self.time()

    ########################
    # FLO
    ########################

    def get_bid(self):
        self.active_run.Flo = Flo(
            params=self.active_run.BidParams, d_graph_id=str(uuid.uuid4())
        )
        slot_start = self.active_run.Slot.StartUnixS
        time_str = pendulum.from_timestamp(slot_start).strftime("%m/%d/%Y, %H:%M")
        # LOGGER.info(f"[{self.time_str()}: {self.short_alias}] Solving Flo ")
        self.active_run.Flo.solve_dijkstra()
        # export_xlsx(alias=self.alias, flo=flo, atn_params=self.atn_params)
        node = self.active_run.Flo.node[0][self.active_run.Flo.starting_store_idx]
        edge: Edge = list(
            filter(
                lambda x: x.end_idx == node.best_next_idx,
                self.active_run.Flo.edges[node],
            )
        )[0]
        q = edge.hp_electricity_avg_kw + edge.boost_electricity_used_avg_kw
        pq = PriceQuantityUnitless(
            PriceTimes1000=int(
                self.active_run.Flo.params.RealtimeElectricityPrice[0] * 1000
            ),
            QuantityTimes1000=int(q * 1000),
        )
        bid_list = [pq]
        self.active_run.Bid.BidList = bid_list

    ########################
    # Market transactions
    ######################

    def pay_market_fee(self) -> str:
        market_fee_micro_algos = 2 * 10**3
        txn = transaction.PaymentTxn(
            sender=self.algo_acct.addr,
            receiver=self.settings.market_maker_algo_address,
            amt=market_fee_micro_algos,
            sp=self.algo_client.suggested_params(),
        )
        signed_txn = txn.sign(self.algo_acct.sk)
        try:
            self.algo_client.send_transaction(signed_txn)
        except Exception as e:
            note = f"Algorand Failure sending market payment transaction: {e}"
            LOGGER.info(note)
            return atn_utils.DUMMY_ALGO_TXN
        try:
            algo_utils.wait_for_transaction(self.algo_client, signed_txn.get_txid())
        except:
            return atn_utils.DUMMY_ALGO_TXN
        return encoding.msgpack_encode(signed_txn)

    def submit_bid(self, slot_stuff: SlotStuff) -> RestfulResponse:
        slot_stuff.Bid.SignedMarketFeeTxn = self.pay_market_fee()
        slot_start = slot_stuff.Slot.StartUnixS
        time_str = pendulum.from_timestamp(slot_start).strftime("%m/%d/%Y, %H:%M")
        # LOGGER.info(
        #     f"[{self.time_str()}: {self.short_alias}] Submitting bid to MarketMaker RestAPI for Slot starting {time_str}"
        # )

        api_endpoint = f"{self.settings.mm_api_root}/atn-bid/"
        r = requests.post(url=api_endpoint, json=slot_stuff.Bid.as_dict())

        # Duplicate on rabbit broker for ear
        self.send_message(
            payload=slot_stuff.Bid,
            to_role=GNodeRole.MarketMaker,
            to_g_node_alias=self.settings.market_maker_alias,
        )

        if r.status_code > 200:
            if r.status_code == 422:
                note = f"Error entering SLA: " + r.json()["detail"]
            else:
                note = r.reason
            return RestfulResponse(Note=note, HttpStatusCode=422)
        else:
            rr = RestfulResponse(**r.json())
            if rr.PayloadTypeName == AcceptedBid_Maker.type_name:
                ab = AcceptedBid_Maker.dict_to_tuple(rr.PayloadAsDict)
                self.send_message(
                    payload=ab, message_category=MessageCategory.RabbitJsonBroadcast
                )
            # pprint(rr)
            return rr

    ########################
    # Initialization
    ########################

    def dev_get_initial_params(self) -> None:
        if not atn_utils.is_dummy_atn_params(self.atn_params):
            LOGGER.warning("Tried to get initial params but already have them")

        now_ms = int(self.time()) * 1000
        self.atn_params = dev_io.atn_params_from_alias(alias=self.alias, now_ms=now_ms)
        self.atn_params.GNodeInstanceId = self.g_node_instance_id
        if atn_utils.is_dummy_atn_params(self.atn_params):
            LOGGER.warning(f"No atn_params for {self.alias} before {self.time_str()}")
            return
        payload = AtnParamsReport_Maker(
            g_node_alias=self.alias,
            g_node_instance_id=self.g_node_instance_id,
            time_unix_s=int(self.time()),
            atn_params=self.atn_params,
            irl_time_unix_s=int(time.time()),
        ).tuple

        self.send_message(
            payload=payload, message_category=MessageCategory.RabbitJsonBroadcast
        )
        # LOGGER.info(f"[{self.time_str()}: {self.short_alias}] Initial params loaded")
        # pprint(self.atn_params)

    def get_initial_params(self) -> None:
        if self.universe_type == UniverseType.Dev:
            self.dev_get_initial_params()
        else:
            raise NotImplementedError

    @property
    def short_alias(self) -> str:
        return self.alias.split(".")[-1]

    ##################
    # Market slot starts
    ##################

    def last_slot(self, market_type: MarketTypeGt) -> MarketSlot:
        start_s = self.active_slot(
            market_type
        ).StartUnixS - self.market_slot_duration_s(market_type)
        return MarketSlot(
            Type=market_type,
            MarketMakerAlias=self.active_slot(market_type).MarketMakerAlias,
            StartUnixS=start_s,
        )

    def active_slot(self, market_type: MarketTypeGt) -> MarketSlot:
        market_delta = market_type.DurationMinutes * 60
        start = math.floor(self.time() / market_delta) * market_delta
        return MarketSlot(
            Type=market_type,
            MarketMakerAlias=self.settings.market_maker_alias,
            StartUnixS=start,
        )

    def next_slot(self, market_type: MarketTypeGt) -> MarketSlot:
        start_s = self.active_slot(
            market_type
        ).StartUnixS + self.market_slot_duration_s(market_type)
        return MarketSlot(
            Type=market_type,
            MarketMakerAlias=self.active_slot(market_type).MarketMakerAlias,
            StartUnixS=start_s,
        )

    def market_slot_duration_s(self, market_type: MarketTypeGt) -> int:
        return market_type.DurationMinutes * 60

    #############
    # TerminalAsset properties
    #################

    def store_idx(self) -> int:
        return round(
            self.atn_params.StorageSteps
            * self.store_kwh
            / atn_utils.get_max_store_kwh_th(self.atn_params)
        )



"""Type atn.params.brickstorageheater, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.enums import DistributionTariff
from gwatn.enums import EnergySupplyType
from gwatn.enums import RecognizedCurrencyUnit
from gwatn.enums import RecognizedTemperatureUnit


class DistributionTariff000SchemaEnum:
    enum_name: str = "distribution.tariff.000"
    symbols: List[str] = [
        "00000000",
        "2127aba6",
        "ea5c675a",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class DistributionTariff000(StrEnum):
    Unknown = auto()
    VersantStorageHeatTariff = auto()
    VersantATariff = auto()

    @classmethod
    def default(cls) -> "DistributionTariff000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class DistributionTariffMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> DistributionTariff:
        if not DistributionTariff000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to DistributionTariff000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, DistributionTariff, DistributionTariff.default())

    @classmethod
    def local_to_type(cls, distribution_tariff: DistributionTariff) -> str:
        if not isinstance(distribution_tariff, DistributionTariff):
            raise SchemaError(
                f"{distribution_tariff} must be of type {DistributionTariff}"
            )
        versioned_enum = as_enum(
            distribution_tariff, DistributionTariff000, DistributionTariff000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, DistributionTariff000] = {
        "00000000": DistributionTariff000.Unknown,
        "2127aba6": DistributionTariff000.VersantStorageHeatTariff,
        "ea5c675a": DistributionTariff000.VersantATariff,
    }

    versioned_enum_to_type_dict: Dict[DistributionTariff000, str] = {
        DistributionTariff000.Unknown: "00000000",
        DistributionTariff000.VersantStorageHeatTariff: "2127aba6",
        DistributionTariff000.VersantATariff: "ea5c675a",
    }


class RecognizedCurrencyUnit000SchemaEnum:
    enum_name: str = "recognized.currency.unit.000"
    symbols: List[str] = [
        "00000000",
        "e57c5143",
        "f7b38fc5",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class RecognizedCurrencyUnit000(StrEnum):
    UNKNOWN = auto()
    USD = auto()
    GBP = auto()

    @classmethod
    def default(cls) -> "RecognizedCurrencyUnit000":
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class RecognizedCurrencyUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> RecognizedCurrencyUnit:
        if not RecognizedCurrencyUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to RecognizedCurrencyUnit000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )

    @classmethod
    def local_to_type(cls, recognized_currency_unit: RecognizedCurrencyUnit) -> str:
        if not isinstance(recognized_currency_unit, RecognizedCurrencyUnit):
            raise SchemaError(
                f"{recognized_currency_unit} must be of type {RecognizedCurrencyUnit}"
            )
        versioned_enum = as_enum(
            recognized_currency_unit,
            RecognizedCurrencyUnit000,
            RecognizedCurrencyUnit000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, RecognizedCurrencyUnit000] = {
        "00000000": RecognizedCurrencyUnit000.UNKNOWN,
        "e57c5143": RecognizedCurrencyUnit000.USD,
        "f7b38fc5": RecognizedCurrencyUnit000.GBP,
    }

    versioned_enum_to_type_dict: Dict[RecognizedCurrencyUnit000, str] = {
        RecognizedCurrencyUnit000.UNKNOWN: "00000000",
        RecognizedCurrencyUnit000.USD: "e57c5143",
        RecognizedCurrencyUnit000.GBP: "f7b38fc5",
    }


class RecognizedTemperatureUnit000SchemaEnum:
    enum_name: str = "recognized.temperature.unit.000"
    symbols: List[str] = [
        "00000000",
        "6f16ee63",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class RecognizedTemperatureUnit000(StrEnum):
    C = auto()
    F = auto()

    @classmethod
    def default(cls) -> "RecognizedTemperatureUnit000":
        return cls.C

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class RecognizedTemperatureUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> RecognizedTemperatureUnit:
        if not RecognizedTemperatureUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to RecognizedTemperatureUnit000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum,
            RecognizedTemperatureUnit,
            RecognizedTemperatureUnit.default(),
        )

    @classmethod
    def local_to_type(
        cls, recognized_temperature_unit: RecognizedTemperatureUnit
    ) -> str:
        if not isinstance(recognized_temperature_unit, RecognizedTemperatureUnit):
            raise SchemaError(
                f"{recognized_temperature_unit} must be of type {RecognizedTemperatureUnit}"
            )
        versioned_enum = as_enum(
            recognized_temperature_unit,
            RecognizedTemperatureUnit000,
            RecognizedTemperatureUnit000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, RecognizedTemperatureUnit000] = {
        "00000000": RecognizedTemperatureUnit000.C,
        "6f16ee63": RecognizedTemperatureUnit000.F,
    }

    versioned_enum_to_type_dict: Dict[RecognizedTemperatureUnit000, str] = {
        RecognizedTemperatureUnit000.C: "00000000",
        RecognizedTemperatureUnit000.F: "6f16ee63",
    }


class EnergySupplyType000SchemaEnum:
    enum_name: str = "energy.supply.type.000"
    symbols: List[str] = [
        "00000000",
        "cb18f937",
        "e9dc99a6",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class EnergySupplyType000(StrEnum):
    Unknown = auto()
    StandardOffer = auto()
    RealtimeLocalLmp = auto()

    @classmethod
    def default(cls) -> "EnergySupplyType000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class EnergySupplyTypeMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> EnergySupplyType:
        if not EnergySupplyType000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to EnergySupplyType000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, EnergySupplyType, EnergySupplyType.default())

    @classmethod
    def local_to_type(cls, energy_supply_type: EnergySupplyType) -> str:
        if not isinstance(energy_supply_type, EnergySupplyType):
            raise SchemaError(
                f"{energy_supply_type} must be of type {EnergySupplyType}"
            )
        versioned_enum = as_enum(
            energy_supply_type, EnergySupplyType000, EnergySupplyType000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, EnergySupplyType000] = {
        "00000000": EnergySupplyType000.Unknown,
        "cb18f937": EnergySupplyType000.StandardOffer,
        "e9dc99a6": EnergySupplyType000.RealtimeLocalLmp,
    }

    versioned_enum_to_type_dict: Dict[EnergySupplyType000, str] = {
        EnergySupplyType000.Unknown: "00000000",
        EnergySupplyType000.StandardOffer: "cb18f937",
        EnergySupplyType000.RealtimeLocalLmp: "e9dc99a6",
    }


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


class AtnParamsBrickstorageheater(BaseModel):
    """Params for a BrickStorageHeater Atn Strategy.

    Parameters for a model of a ceramic brick thermal storage room unit, using a two-dimensional Dijstra graph parameterized by 'fullness of the thermal battery' and time.
    [More info](https://gridworks-atn.readthedocs.io/en/latest/brick-storage-heater.html).
    """

    GNodeAlias: str = Field(
        title="GNode Alias",
    )
    HomeCity: str = Field(
        title="Home City",
        default="MILLINOCKET_ME",
    )
    TimezoneString: str = Field(
        title="Timezone String",
        default="US/Eastern",
    )
    StorageSteps: int = Field(
        title="Storage Steps",
        description="The number of nodes in each TimeSlice for the Dijkstra graph.",
        default=100,
    )
    FloSlices: int = Field(
        title="FloSlices",
        description="The number of nodes for each storage level in the Dijstra graph.",
        default=48,
    )
    SliceDurationMinutes: int = Field(
        title="SliceDurationMinutes",
        description="The length of",
        default=60,
    )
    CurrencyUnit: RecognizedCurrencyUnit = Field(
        title="CurrencyUnit",
        default=RecognizedCurrencyUnit.USD,
    )
    Tariff: DistributionTariff = Field(
        title="Tariff",
        default=DistributionTariff.VersantStorageHeatTariff,
    )
    EnergyType: EnergySupplyType = Field(
        title="EnergyType",
        default=EnergySupplyType.RealtimeLocalLmp,
    )
    StandardOfferPriceDollarsPerMwh: int = Field(
        title="StandardOfferPriceDollarsPerMwh",
        default=110,
    )
    DistributionTariffDollarsPerMwh: int = Field(
        title="DistributionTariffDollarsPerMwh",
        default=113,
    )
    MaxBrickTempC: int = Field(
        title="MaxBrickTempC",
        default=190,
    )
    RatedMaxPowerKw: float = Field(
        title="RatedMaxPowerKw",
        default=13.5,
    )
    C: float = Field(
        title="C",
        default=200,
    )
    ROff: float = Field(
        title="ROff",
        default=0.08,
    )
    ROn: float = Field(
        title="ROn",
        default=0.15,
    )
    RoomTempF: int = Field(
        title="RoomTempF",
        default=70,
    )
    AnnualHvacKwhTh: int = Field(
        title="AnnualHvacKwhTh",
        default=28125,
    )
    BetaOt: int = Field(
        title="BetaOt",
        default=158,
    )
    TempUnit: RecognizedTemperatureUnit = Field(
        title="TempUnit",
        default=RecognizedTemperatureUnit.F,
    )
    AmbientPowerInKw: float = Field(
        title="AmbientPowerInKw",
        default=1.25,
    )
    HouseWorstCaseTempF: int = Field(
        title="HouseWorstCaseTempF",
        default=-7,
    )
    TypeName: Literal["atn.params.brickstorageheater"] = "atn.params.brickstorageheater"
    Version: str = "000"

    @validator("GNodeAlias")
    def _check_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"GNodeAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("CurrencyUnit")
    def _check_currency_unit(cls, v: RecognizedCurrencyUnit) -> RecognizedCurrencyUnit:
        return as_enum(v, RecognizedCurrencyUnit, RecognizedCurrencyUnit.UNKNOWN)

    @validator("Tariff")
    def _check_tariff(cls, v: DistributionTariff) -> DistributionTariff:
        return as_enum(v, DistributionTariff, DistributionTariff.Unknown)

    @validator("EnergyType")
    def _check_energy_type(cls, v: EnergySupplyType) -> EnergySupplyType:
        return as_enum(v, EnergySupplyType, EnergySupplyType.Unknown)

    @validator("TempUnit")
    def _check_temp_unit(
        cls, v: RecognizedTemperatureUnit
    ) -> RecognizedTemperatureUnit:
        return as_enum(v, RecognizedTemperatureUnit, RecognizedTemperatureUnit.C)

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["CurrencyUnit"]
        CurrencyUnit = as_enum(
            self.CurrencyUnit, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )
        d["CurrencyUnitGtEnumSymbol"] = RecognizedCurrencyUnitMap.local_to_type(
            CurrencyUnit
        )
        del d["Tariff"]
        Tariff = as_enum(self.Tariff, DistributionTariff, DistributionTariff.default())
        d["TariffGtEnumSymbol"] = DistributionTariffMap.local_to_type(Tariff)
        del d["EnergyType"]
        EnergyType = as_enum(
            self.EnergyType, EnergySupplyType, EnergySupplyType.default()
        )
        d["EnergyTypeGtEnumSymbol"] = EnergySupplyTypeMap.local_to_type(EnergyType)
        del d["TempUnit"]
        TempUnit = as_enum(
            self.TempUnit,
            RecognizedTemperatureUnit,
            RecognizedTemperatureUnit.default(),
        )
        d["TempUnitGtEnumSymbol"] = RecognizedTemperatureUnitMap.local_to_type(TempUnit)
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class AtnParamsBrickstorageheater_Maker:
    type_name = "atn.params.brickstorageheater"
    version = "000"

    def __init__(
        self,
        g_node_alias: str,
        home_city: str,
        timezone_string: str,
        storage_steps: int,
        flo_slices: int,
        slice_duration_minutes: int,
        currency_unit: RecognizedCurrencyUnit,
        tariff: DistributionTariff,
        energy_type: EnergySupplyType,
        standard_offer_price_dollars_per_mwh: int,
        distribution_tariff_dollars_per_mwh: int,
        max_brick_temp_c: int,
        rated_max_power_kw: float,
        c: float,
        r_off: float,
        r_on: float,
        room_temp_f: int,
        annual_hvac_kwh_th: int,
        beta_ot: int,
        temp_unit: RecognizedTemperatureUnit,
        ambient_power_in_kw: float,
        house_worst_case_temp_f: int,
    ):
        self.tuple = AtnParamsBrickstorageheater(
            GNodeAlias=g_node_alias,
            HomeCity=home_city,
            TimezoneString=timezone_string,
            StorageSteps=storage_steps,
            FloSlices=flo_slices,
            SliceDurationMinutes=slice_duration_minutes,
            CurrencyUnit=currency_unit,
            Tariff=tariff,
            EnergyType=energy_type,
            StandardOfferPriceDollarsPerMwh=standard_offer_price_dollars_per_mwh,
            DistributionTariffDollarsPerMwh=distribution_tariff_dollars_per_mwh,
            MaxBrickTempC=max_brick_temp_c,
            RatedMaxPowerKw=rated_max_power_kw,
            C=c,
            ROff=r_off,
            ROn=r_on,
            RoomTempF=room_temp_f,
            AnnualHvacKwhTh=annual_hvac_kwh_th,
            BetaOt=beta_ot,
            TempUnit=temp_unit,
            AmbientPowerInKw=ambient_power_in_kw,
            HouseWorstCaseTempF=house_worst_case_temp_f,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: AtnParamsBrickstorageheater) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> AtnParamsBrickstorageheater:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> AtnParamsBrickstorageheater:
        d2 = dict(d)
        if "GNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeAlias")
        if "HomeCity" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HomeCity")
        if "TimezoneString" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TimezoneString")
        if "StorageSteps" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StorageSteps")
        if "FloSlices" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FloSlices")
        if "SliceDurationMinutes" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SliceDurationMinutes")
        if "CurrencyUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing CurrencyUnitGtEnumSymbol")
        if (
            d2["CurrencyUnitGtEnumSymbol"]
            in RecognizedCurrencyUnit000SchemaEnum.symbols
        ):
            d2["CurrencyUnit"] = RecognizedCurrencyUnitMap.type_to_local(
                d2["CurrencyUnitGtEnumSymbol"]
            )
        else:
            d2["CurrencyUnit"] = RecognizedCurrencyUnit.default()
        if "TariffGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TariffGtEnumSymbol")
        if d2["TariffGtEnumSymbol"] in DistributionTariff000SchemaEnum.symbols:
            d2["Tariff"] = DistributionTariffMap.type_to_local(d2["TariffGtEnumSymbol"])
        else:
            d2["Tariff"] = DistributionTariff.default()
        if "EnergyTypeGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing EnergyTypeGtEnumSymbol")
        if d2["EnergyTypeGtEnumSymbol"] in EnergySupplyType000SchemaEnum.symbols:
            d2["EnergyType"] = EnergySupplyTypeMap.type_to_local(
                d2["EnergyTypeGtEnumSymbol"]
            )
        else:
            d2["EnergyType"] = EnergySupplyType.default()
        if "StandardOfferPriceDollarsPerMwh" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StandardOfferPriceDollarsPerMwh")
        if "DistributionTariffDollarsPerMwh" not in d2.keys():
            raise SchemaError(f"dict {d2} missing DistributionTariffDollarsPerMwh")
        if "MaxBrickTempC" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MaxBrickTempC")
        if "RatedMaxPowerKw" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RatedMaxPowerKw")
        if "C" not in d2.keys():
            raise SchemaError(f"dict {d2} missing C")
        if "ROff" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ROff")
        if "ROn" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ROn")
        if "RoomTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RoomTempF")
        if "AnnualHvacKwhTh" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AnnualHvacKwhTh")
        if "BetaOt" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BetaOt")
        if "TempUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TempUnitGtEnumSymbol")
        if d2["TempUnitGtEnumSymbol"] in RecognizedTemperatureUnit000SchemaEnum.symbols:
            d2["TempUnit"] = RecognizedTemperatureUnitMap.type_to_local(
                d2["TempUnitGtEnumSymbol"]
            )
        else:
            d2["TempUnit"] = RecognizedTemperatureUnit.default()
        if "AmbientPowerInKw" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AmbientPowerInKw")
        if "HouseWorstCaseTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HouseWorstCaseTempF")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return AtnParamsBrickstorageheater(
            GNodeAlias=d2["GNodeAlias"],
            HomeCity=d2["HomeCity"],
            TimezoneString=d2["TimezoneString"],
            StorageSteps=d2["StorageSteps"],
            FloSlices=d2["FloSlices"],
            SliceDurationMinutes=d2["SliceDurationMinutes"],
            CurrencyUnit=d2["CurrencyUnit"],
            Tariff=d2["Tariff"],
            EnergyType=d2["EnergyType"],
            StandardOfferPriceDollarsPerMwh=d2["StandardOfferPriceDollarsPerMwh"],
            DistributionTariffDollarsPerMwh=d2["DistributionTariffDollarsPerMwh"],
            MaxBrickTempC=d2["MaxBrickTempC"],
            RatedMaxPowerKw=d2["RatedMaxPowerKw"],
            C=d2["C"],
            ROff=d2["ROff"],
            ROn=d2["ROn"],
            RoomTempF=d2["RoomTempF"],
            AnnualHvacKwhTh=d2["AnnualHvacKwhTh"],
            BetaOt=d2["BetaOt"],
            TempUnit=d2["TempUnit"],
            AmbientPowerInKw=d2["AmbientPowerInKw"],
            HouseWorstCaseTempF=d2["HouseWorstCaseTempF"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type initial.tadeed.algo.optin, version 002"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class InitialTadeedAlgoOptin(BaseModel):
    """Received by TaDaemon so that it can opt into intial TaDeed.

    The TaDaemon must opt into the TaDeed before receiving it. This message prompts that action.
    """

    TerminalAssetAlias: str = Field(
        title="The GNodeAlias of the TerminalAsset",
    )
    TaOwnerAddr: str = Field(
        title="The Algorand address of the owner for the TerminalAsset",
    )
    ValidatorAddr: str = Field(
        title="Address of the TaValidator",
        description="The Algorand address of the TaValidator who has validated the location, device type, and power metering of the TerminalAsset.",
    )
    SignedInitialDaemonFundingTxn: str = Field(
        title="SignedInitialDaemonFundingTxn",
        description="Funding transaction for the TaDaemon account, signed by the TaOwner.",
    )
    TypeName: Literal["initial.tadeed.algo.optin"] = "initial.tadeed.algo.optin"
    Version: str = "002"

    @validator("TerminalAssetAlias")
    def _check_terminal_asset_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"TerminalAssetAlias failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("TaOwnerAddr")
    def _check_ta_owner_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"TaOwnerAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("ValidatorAddr")
    def _check_validator_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("SignedInitialDaemonFundingTxn")
    def _check_signed_initial_daemon_funding_txn(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"SignedInitialDaemonFundingTxn failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator(pre=True)
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Is correct Multisig.
        Decoded SignedInitialDaemonFundingTxn must be a SignedTransaction signed by TaOwnerAddr.
        """
        # TODO: Implement check for axiom 1"
        return v

    @root_validator
    def check_axiom_2(cls, v: dict) -> dict:
        """
        Axiom 2: TaDeed consistency.
        There is an ASA TaDeed created by and owned by the 2-sig MultiAccount [GnfAdminAddr, ValidatorAddr],
        where the TaDeed's AssetName is equal to the payload's TerminalAssetAlias.
        """
        # TODO: Implement check for axiom 2"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class InitialTadeedAlgoOptin_Maker:
    type_name = "initial.tadeed.algo.optin"
    version = "002"

    def __init__(
        self,
        terminal_asset_alias: str,
        ta_owner_addr: str,
        validator_addr: str,
        signed_initial_daemon_funding_txn: str,
    ):
        self.tuple = InitialTadeedAlgoOptin(
            TerminalAssetAlias=terminal_asset_alias,
            TaOwnerAddr=ta_owner_addr,
            ValidatorAddr=validator_addr,
            SignedInitialDaemonFundingTxn=signed_initial_daemon_funding_txn,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: InitialTadeedAlgoOptin) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> InitialTadeedAlgoOptin:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> InitialTadeedAlgoOptin:
        d2 = dict(d)
        if "TerminalAssetAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TerminalAssetAlias")
        if "TaOwnerAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TaOwnerAddr")
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "SignedInitialDaemonFundingTxn" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SignedInitialDaemonFundingTxn")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return InitialTadeedAlgoOptin(
            TerminalAssetAlias=d2["TerminalAssetAlias"],
            TaOwnerAddr=d2["TaOwnerAddr"],
            ValidatorAddr=d2["ValidatorAddr"],
            SignedInitialDaemonFundingTxn=d2["SignedInitialDaemonFundingTxn"],
            TypeName=d2["TypeName"],
            Version="002",
        )



from gridworks.data_classes.base_g_node import BaseGNode
from gridworks.data_classes.g_node import GNode
from gridworks.data_classes.g_node_instance import GNodeInstance
from gridworks.data_classes.gps_point import GpsPoint
from gridworks.data_classes.market_type import MarketType
from gridworks.data_classes.supervisor_container import SupervisorContainer

from gwatn.data_classes.d_edge import DEdge
from gwatn.data_classes.d_node import DNode


__all__ = [
    "BaseGNode",
    "DEdge",
    "DNode",
    "GNode",
    "GNodeInstance",
    "GpsPoint",
    "MarketType",
    "SupervisorContainer",
]



import json
from typing import Literal
from typing import Optional

from gwproto.enums import TelemetryName as EnumTelemetryName
from gwproto.errors import MpSchemaError
from gwproto.message import as_enum
from pydantic import BaseModel
from pydantic import Field


class DataChannel(BaseModel):
    """Data Channel.

    A data channel is a concept of some collection of readings that share all characteristics other than time.
    """

    DisplayName: str = Field(
        title="Display Name",
    )
    AboutName: str = Field(
        title="AboutName",
    )
    FromName: str = Field(
        title="FromName",
    )

    TelemetryName: EnumTelemetryName = Field(
        title="TelemetryName",
    )

    ExpectedMaxValue: Optional[int] = Field(title="Expected Max Value, if appropriate")

    ExpectedMinValue: Optional[int] = Field(title="Expected Min Value, if appropriate")

    TypeName: Literal["data.channel"] = "data.channel"
    Version: str = "000"

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa



""" List of all the types """

# From gridworks
from gridworks.types import BaseGNodeGt
from gridworks.types import BaseGNodeGt_Maker
from gridworks.types import GNodeGt
from gridworks.types import GNodeGt_Maker
from gridworks.types import GNodeInstanceGt
from gridworks.types import GNodeInstanceGt_Maker
from gridworks.types import GwCertId
from gridworks.types import GwCertId_Maker
from gridworks.types import HeartbeatA
from gridworks.types import HeartbeatA_Maker
from gridworks.types import Ready
from gridworks.types import Ready_Maker
from gridworks.types import SimTimestep
from gridworks.types import SimTimestep_Maker
from gridworks.types import SuperStarter
from gridworks.types import SuperStarter_Maker
from gridworks.types import SupervisorContainerGt
from gridworks.types import SupervisorContainerGt_Maker

# From gwproto
from gwproto.types import ComponentAttributeClassGt
from gwproto.types import ComponentAttributeClassGt_Maker
from gwproto.types import ComponentGt
from gwproto.types import ComponentGt_Maker
from gwproto.types import DataChannel
from gwproto.types import DataChannel_Maker
from gwproto.types import EgaugeIo
from gwproto.types import EgaugeIo_Maker
from gwproto.types import EgaugeRegisterConfig
from gwproto.types import EgaugeRegisterConfig_Maker
from gwproto.types import ElectricMeterCacGt
from gwproto.types import ElectricMeterCacGt_Maker
from gwproto.types import GtDispatchBoolean
from gwproto.types import GtDispatchBoolean_Maker
from gwproto.types import GtDispatchBooleanLocal
from gwproto.types import GtDispatchBooleanLocal_Maker
from gwproto.types import GtDriverBooleanactuatorCmd
from gwproto.types import GtDriverBooleanactuatorCmd_Maker
from gwproto.types import GtShBooleanactuatorCmdStatus
from gwproto.types import GtShBooleanactuatorCmdStatus_Maker
from gwproto.types import GtShCliAtnCmd
from gwproto.types import GtShCliAtnCmd_Maker
from gwproto.types import GtShMultipurposeTelemetryStatus
from gwproto.types import GtShMultipurposeTelemetryStatus_Maker
from gwproto.types import GtShSimpleTelemetryStatus
from gwproto.types import GtShSimpleTelemetryStatus_Maker
from gwproto.types import GtShStatus
from gwproto.types import GtShStatus_Maker
from gwproto.types import GtShTelemetryFromMultipurposeSensor
from gwproto.types import GtShTelemetryFromMultipurposeSensor_Maker
from gwproto.types import GtTelemetry
from gwproto.types import GtTelemetry_Maker
from gwproto.types import HeartbeatB
from gwproto.types import HeartbeatB_Maker
from gwproto.types import MultipurposeSensorCacGt
from gwproto.types import MultipurposeSensorCacGt_Maker
from gwproto.types import PipeFlowSensorCacGt
from gwproto.types import PipeFlowSensorCacGt_Maker
from gwproto.types import PipeFlowSensorComponentGt
from gwproto.types import PipeFlowSensorComponentGt_Maker
from gwproto.types import PowerWatts
from gwproto.types import PowerWatts_Maker
from gwproto.types import RelayCacGt
from gwproto.types import RelayCacGt_Maker
from gwproto.types import RelayComponentGt
from gwproto.types import RelayComponentGt_Maker
from gwproto.types import ResistiveHeaterCacGt
from gwproto.types import ResistiveHeaterCacGt_Maker
from gwproto.types import ResistiveHeaterComponentGt
from gwproto.types import ResistiveHeaterComponentGt_Maker
from gwproto.types import SimpleTempSensorCacGt
from gwproto.types import SimpleTempSensorCacGt_Maker
from gwproto.types import SimpleTempSensorComponentGt
from gwproto.types import SimpleTempSensorComponentGt_Maker
from gwproto.types import SnapshotSpaceheat
from gwproto.types import SnapshotSpaceheat_Maker
from gwproto.types import SpaceheatNodeGt
from gwproto.types import SpaceheatNodeGt_Maker
from gwproto.types import TelemetryReportingConfig
from gwproto.types import TelemetryReportingConfig_Maker
from gwproto.types import TelemetrySnapshotSpaceheat
from gwproto.types import TelemetrySnapshotSpaceheat_Maker
from gwproto.types.electric_meter_component_gt import ElectricMeterComponentGt
from gwproto.types.electric_meter_component_gt import ElectricMeterComponentGt_Maker
from gwproto.types.multipurpose_sensor_component_gt import MultipurposeSensorComponentGt
from gwproto.types.multipurpose_sensor_component_gt import (
    MultipurposeSensorComponentGt_Maker,
)

# From gwatn
from gwatn.types.accepted_bid import AcceptedBid
from gwatn.types.accepted_bid import AcceptedBid_Maker
from gwatn.types.atn_bid import AtnBid
from gwatn.types.atn_bid import AtnBid_Maker
from gwatn.types.atn_params import AtnParams
from gwatn.types.atn_params import AtnParams_Maker
from gwatn.types.atn_params_brickstorageheater import AtnParamsBrickstorageheater
from gwatn.types.atn_params_brickstorageheater import AtnParamsBrickstorageheater_Maker
from gwatn.types.atn_params_report import AtnParamsReport
from gwatn.types.atn_params_report import AtnParamsReport_Maker
from gwatn.types.basegnode_scada_create import BasegnodeScadaCreate
from gwatn.types.basegnode_scada_create import BasegnodeScadaCreate_Maker
from gwatn.types.discoverycert_algo_create import DiscoverycertAlgoCreate
from gwatn.types.discoverycert_algo_create import DiscoverycertAlgoCreate_Maker
from gwatn.types.dispatch_contract_confirmed import DispatchContractConfirmed
from gwatn.types.dispatch_contract_confirmed import DispatchContractConfirmed_Maker
from gwatn.types.flo_params import FloParams
from gwatn.types.flo_params import FloParams_Maker
from gwatn.types.flo_params_brickstorageheater import FloParamsBrickstorageheater
from gwatn.types.flo_params_brickstorageheater import FloParamsBrickstorageheater_Maker
from gwatn.types.flo_params_report import FloParamsReport
from gwatn.types.flo_params_report import FloParamsReport_Maker
from gwatn.types.initial_tadeed_algo_create import InitialTadeedAlgoCreate
from gwatn.types.initial_tadeed_algo_create import InitialTadeedAlgoCreate_Maker
from gwatn.types.initial_tadeed_algo_optin import InitialTadeedAlgoOptin
from gwatn.types.initial_tadeed_algo_optin import InitialTadeedAlgoOptin_Maker
from gwatn.types.initial_tadeed_algo_transfer import InitialTadeedAlgoTransfer
from gwatn.types.initial_tadeed_algo_transfer import InitialTadeedAlgoTransfer_Maker
from gwatn.types.join_dispatch_contract import JoinDispatchContract
from gwatn.types.join_dispatch_contract import JoinDispatchContract_Maker
from gwatn.types.latest_price import LatestPrice
from gwatn.types.latest_price import LatestPrice_Maker
from gwatn.types.market_slot import MarketSlot
from gwatn.types.market_slot import MarketSlot_Maker
from gwatn.types.market_type_gt import MarketTypeGt
from gwatn.types.market_type_gt import MarketTypeGt_Maker
from gwatn.types.new_tadeed_algo_optin import NewTadeedAlgoOptin
from gwatn.types.new_tadeed_algo_optin import NewTadeedAlgoOptin_Maker
from gwatn.types.new_tadeed_send import NewTadeedSend
from gwatn.types.new_tadeed_send import NewTadeedSend_Maker
from gwatn.types.old_tadeed_algo_return import OldTadeedAlgoReturn
from gwatn.types.old_tadeed_algo_return import OldTadeedAlgoReturn_Maker
from gwatn.types.price_quantity import PriceQuantity
from gwatn.types.price_quantity import PriceQuantity_Maker
from gwatn.types.price_quantity_unitless import PriceQuantityUnitless
from gwatn.types.price_quantity_unitless import PriceQuantityUnitless_Maker
from gwatn.types.scada_cert_transfer import ScadaCertTransfer
from gwatn.types.scada_cert_transfer import ScadaCertTransfer_Maker
from gwatn.types.simplesim_driver_data import SimplesimDriverData
from gwatn.types.simplesim_driver_data import SimplesimDriverData_Maker
from gwatn.types.simplesim_driver_data_bsh import SimplesimDriverDataBsh
from gwatn.types.simplesim_driver_data_bsh import SimplesimDriverDataBsh_Maker
from gwatn.types.simplesim_driver_report import SimplesimDriverReport
from gwatn.types.simplesim_driver_report import SimplesimDriverReport_Maker
from gwatn.types.simplesim_snapshot_brickstorageheater import (
    SimplesimSnapshotBrickstorageheater,
)
from gwatn.types.simplesim_snapshot_brickstorageheater import (
    SimplesimSnapshotBrickstorageheater_Maker,
)
from gwatn.types.sla_enter import SlaEnter
from gwatn.types.sla_enter import SlaEnter_Maker
from gwatn.types.tadeed_specs_hack import TadeedSpecsHack
from gwatn.types.tadeed_specs_hack import TadeedSpecsHack_Maker
from gwatn.types.tavalidatorcert_algo_create import TavalidatorcertAlgoCreate
from gwatn.types.tavalidatorcert_algo_create import TavalidatorcertAlgoCreate_Maker
from gwatn.types.tavalidatorcert_algo_transfer import TavalidatorcertAlgoTransfer
from gwatn.types.tavalidatorcert_algo_transfer import TavalidatorcertAlgoTransfer_Maker
from gwatn.types.terminalasset_certify_hack import TerminalassetCertifyHack
from gwatn.types.terminalasset_certify_hack import TerminalassetCertifyHack_Maker


__all__ = [
    "AcceptedBid",
    "AcceptedBid_Maker",
    "AtnBid",
    "AtnBid_Maker",
    "AtnParams",
    "AtnParams_Maker",
    "AtnParamsBrickstorageheater",
    "AtnParamsBrickstorageheater_Maker",
    "AtnParamsReport",
    "AtnParamsReport_Maker",
    "BaseGNodeGt",
    "BaseGNodeGt_Maker",
    "BasegnodeScadaCreate",
    "BasegnodeScadaCreate_Maker",
    "ComponentAttributeClassGt",
    "ComponentAttributeClassGt_Maker",
    "ComponentGt",
    "ComponentGt_Maker",
    "DataChannel",
    "DataChannel_Maker",
    "DiscoverycertAlgoCreate",
    "DiscoverycertAlgoCreate_Maker",
    "DispatchContractConfirmed",
    "DispatchContractConfirmed_Maker",
    "EgaugeIo",
    "EgaugeIo_Maker",
    "EgaugeRegisterConfig",
    "EgaugeRegisterConfig_Maker",
    "ElectricMeterCacGt",
    "ElectricMeterCacGt_Maker",
    "ElectricMeterComponentGt",
    "ElectricMeterComponentGt_Maker",
    "FloParams",
    "FloParams_Maker",
    "FloParamsBrickstorageheater",
    "FloParamsBrickstorageheater_Maker",
    "FloParamsReport",
    "FloParamsReport_Maker",
    "GNodeGt",
    "GNodeGt_Maker",
    "GNodeInstanceGt",
    "GNodeInstanceGt_Maker",
    "GtDispatchBoolean",
    "GtDispatchBoolean_Maker",
    "GtDispatchBooleanLocal",
    "GtDispatchBooleanLocal_Maker",
    "GtDriverBooleanactuatorCmd",
    "GtDriverBooleanactuatorCmd_Maker",
    "GtShBooleanactuatorCmdStatus",
    "GtShBooleanactuatorCmdStatus_Maker",
    "GtShCliAtnCmd",
    "GtShCliAtnCmd_Maker",
    "GtShMultipurposeTelemetryStatus",
    "GtShMultipurposeTelemetryStatus_Maker",
    "GtShSimpleTelemetryStatus",
    "GtShSimpleTelemetryStatus_Maker",
    "GtShStatus",
    "GtShStatus_Maker",
    "GtShTelemetryFromMultipurposeSensor",
    "GtShTelemetryFromMultipurposeSensor_Maker",
    "GtTelemetry",
    "GtTelemetry_Maker",
    "GwCertId",
    "GwCertId_Maker",
    "HeartbeatA",
    "HeartbeatA_Maker",
    "HeartbeatB",
    "HeartbeatB_Maker",
    "InitialTadeedAlgoCreate",
    "InitialTadeedAlgoCreate_Maker",
    "InitialTadeedAlgoOptin",
    "InitialTadeedAlgoOptin_Maker",
    "InitialTadeedAlgoTransfer",
    "InitialTadeedAlgoTransfer_Maker",
    "JoinDispatchContract",
    "JoinDispatchContract_Maker",
    "LatestPrice",
    "LatestPrice_Maker",
    "MarketSlot",
    "MarketSlot_Maker",
    "MarketTypeGt",
    "MarketTypeGt_Maker",
    "MultipurposeSensorCacGt",
    "MultipurposeSensorCacGt_Maker",
    "MultipurposeSensorComponentGt",
    "MultipurposeSensorComponentGt_Maker",
    "NewTadeedAlgoOptin",
    "NewTadeedAlgoOptin_Maker",
    "NewTadeedSend",
    "NewTadeedSend_Maker",
    "OldTadeedAlgoReturn",
    "OldTadeedAlgoReturn_Maker",
    "PipeFlowSensorCacGt",
    "PipeFlowSensorCacGt_Maker",
    "PipeFlowSensorComponentGt",
    "PipeFlowSensorComponentGt_Maker",
    "PowerWatts",
    "PowerWatts_Maker",
    "PriceQuantity",
    "PriceQuantity_Maker",
    "PriceQuantityUnitless",
    "PriceQuantityUnitless_Maker",
    "Ready",
    "Ready_Maker",
    "RelayCacGt",
    "RelayCacGt_Maker",
    "RelayComponentGt",
    "RelayComponentGt_Maker",
    "ResistiveHeaterCacGt",
    "ResistiveHeaterCacGt_Maker",
    "ResistiveHeaterComponentGt",
    "ResistiveHeaterComponentGt_Maker",
    "ScadaCertTransfer",
    "ScadaCertTransfer_Maker",
    "SimTimestep",
    "SimTimestep_Maker",
    "SimpleTempSensorCacGt",
    "SimpleTempSensorCacGt_Maker",
    "SimpleTempSensorComponentGt",
    "SimpleTempSensorComponentGt_Maker",
    "SimplesimDriverData",
    "SimplesimDriverData_Maker",
    "SimplesimDriverDataBsh",
    "SimplesimDriverDataBsh_Maker",
    "SimplesimDriverReport",
    "SimplesimDriverReport_Maker",
    "SimplesimSnapshotBrickstorageheater",
    "SimplesimSnapshotBrickstorageheater_Maker",
    "SlaEnter",
    "SlaEnter_Maker",
    "SnapshotSpaceheat",
    "SnapshotSpaceheat_Maker",
    "SpaceheatNodeGt",
    "SpaceheatNodeGt_Maker",
    "SuperStarter",
    "SuperStarter_Maker",
    "SupervisorContainerGt",
    "SupervisorContainerGt_Maker",
    "TadeedSpecsHack",
    "TadeedSpecsHack_Maker",
    "TavalidatorcertAlgoCreate",
    "TavalidatorcertAlgoCreate_Maker",
    "TavalidatorcertAlgoTransfer",
    "TavalidatorcertAlgoTransfer_Maker",
    "TelemetryReportingConfig",
    "TelemetryReportingConfig_Maker",
    "TelemetrySnapshotSpaceheat",
    "TelemetrySnapshotSpaceheat_Maker",
    "TerminalassetCertifyHack",
    "TerminalassetCertifyHack_Maker",
]



from typing import Final
from typing import Literal

from beaker import Application
from beaker import ApplicationStateValue
from beaker import Authorize
from beaker import close_out
from beaker import create
from beaker import delete
from beaker import external
from beaker import opt_in
from beaker.lib.storage import Mapping
from pyteal import Approve
from pyteal import Assert
from pyteal import AssetHolding
from pyteal import AssetParam
from pyteal import Bytes
from pyteal import Concat
from pyteal import Expr
from pyteal import Global
from pyteal import Int
from pyteal import Seq
from pyteal import TealType
from pyteal import abi


BoxFlatMinBalance = 2500
BoxByteMinBalance = 400


def commented_assert(conditions: list[tuple[Expr, str]]) -> list[Expr]:
    return [Assert(cond, comment=cmt) for cond, cmt in conditions]


class ScadaCertErrors:
    WrongUnitName = "UnitName not SCADA"
    WrongTotal = "Total not 1"
    WrongDecimals = "Decimals not 0"
    WrongManager = "Manager not GnfAdminAddr"
    WrongCreator = "Creator not GnfAdminAddr"


class AtnCertErrors:
    WrongUnitName = "UnitName not TATRADE"
    WrongTotal = "Total not 1"
    WrongDecimals = "Decimals not 0"
    WrongManager = "Manager not GnfAdminAddr"
    WrongCreator = "Creator not GnfAdminAddr"


class DispatchContractErrors:
    ReceiverNotAppAddr = "receiver not app address"
    AmountLessThanMinimum = "amount minimum not met"
    CreatorNotScada = "creator does not own a ScadaCert"
    AtnDoesNotOwnTradingRights = "Atn does not own the correct TaTradingRights"


class HeartbeatB(abi.NamedTuple):
    """Matches GridWorks type heartbeat.b"""

    FromGNodeAlias: abi.Field[abi.String]
    FromGNodeInstanceId: abi.Field[abi.String]
    MyHex: abi.Field[abi.String]
    YourLastHex: abi.Field[abi.String]
    LastReceivedTimeUnixMs: abi.Field[abi.Uint64]
    SendTimeUnixMs: abi.Field[abi.Uint64]
    TypeName: abi.Field[abi.String] = Literal["heartbeat.b"]
    Version: abi.Field[abi.String] = Literal["000"]


class HeartbeatStorageData(abi.NamedTuple):
    """20 bytes of HB Data. Staticly typed for putting into a box"""

    FromScada: abi.Field[abi.Bool]
    MyHex: abi.Field[abi.Byte]
    YourLastHex: abi.Field[abi.Byte]
    LastReceivedTimeUnixMs: abi.Field[abi.Uint64]  # 8 bytes
    SendTimeUnixMs: abi.Field[abi.Uint64]  # 8 bytes


class DispatchContract(Application):
    """The DispatchContract between an AtomicTNode-Scada pair that is coordinating
    the control of their shared TerminalAsset.

    https://gridworks.readthedocs.io/en/latest/dispatch-contract.html
    """

    RecentHeartbeats = Mapping(abi.Uint8, HeartbeatStorageData)
    # RecentHeartbeats = Mapping(abi.StaticBytes[4], HeartbeatData) <- did not work
    # Two rolling hours of minutely HeartbeatData for both Atn and SCADA
    # SCADA keys tagged 0 through 119, Atn keys tagged 120 through 239
    # where the number is minute of day mod 120
    # _heartbeat_box_balance = 4_000_000
    # 2 months of boolean TalkingWith state change, where
    # _talking_with_audit_box_balance = 6_000_000
    # min_balance = _heartbeat_box_balance + _talking_with_audit_box_balance

    min_balance = 10_000_000
    MinimumBalance = Int(min_balance)

    governor: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.bytes,
        default=Global.creator_address(),
        key=Bytes("g"),
        static=True,
        descr="The governor of this contract",
    )

    ta_alias: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.bytes,
        static=True,
        descr="TerminalAsset GNodeAlias",
    )

    scada_cert_idx: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        static=True,
        descr="The asset id of the Scada Cert ASA",
    )

    ta_trading_rights_idx: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        static=True,
        descr="The asset id of the TaTradingRights ASA",
    )

    @external
    def hello(self, name: abi.String, *, output: abi.String):
        """This is in one of the intro beaker vignettes from their
        great set of examples.

        https://github.com/algorand-devrel/beaker/blob/master/examples/simple/hello.py#L8

        Args:
            name: a string

        Returns:
            Hello, name
        """
        return output.set(Concat(Bytes("Hello, "), name.get()))

    @external(authorize=Authorize.only(governor))
    def bootstrap1(
        self,
        scada_seed: abi.PaymentTransaction,
        ScadaCert: abi.Asset,
        *,
        output: abi.String,
    ):
        """First half of the bootstrap. Checks that
           - ScadaCert is valid
           - payment sender (also the governor) owns the ScadaCert, thus proving their
           identity as Scada

            If these checks pass, then the global scada_addr and ta_alias are set.

        Args:
            scada_seed: Initial Payment transaction to the app, by Scada Acct
            ScadaCert: The asset owned by the Scada Addr validating that the GNodeFactory
            authorized its GNode as a SCADA.

        Returns:
            The TerminalAssetAlias (from the ScadaCert)
        """

        return Seq(
            Assert(
                scada_seed.get().receiver() == self.address,
                comment=DispatchContractErrors.ReceiverNotAppAddr,
            ),
            Assert(
                scada_seed.get().amount() >= self.MinimumBalance,
                comment=DispatchContractErrors.AmountLessThanMinimum,
            ),
            cert_type := AssetParam.unitName(ScadaCert.asset_id()),
            Assert(cert_type.hasValue()),
            Assert(
                cert_type.value() == Bytes("SCADA"),
                comment=ScadaCertErrors.WrongUnitName,
            ),
            ta_alias := AssetParam.name(ScadaCert.asset_id()),
            Assert(ta_alias.hasValue()),
            scada_cert_balance := AssetHolding.balance(
                scada_seed.get().sender(), ScadaCert.asset_id()
            ),
            Assert(scada_cert_balance.hasValue()),
            Assert(
                scada_cert_balance.value() == Int(1),
                comment=DispatchContractErrors.CreatorNotScada,
            ),
            # creator := AssetParam.creator(ScadaCert.asset_id()),
            # Assert(creator.hasValue()),
            # Assert(
            #     creator.value() == config.Public().gnf_admin_addr,
            #     comment=ScadaCertErrors.WrongCreator,
            # ),
            # manager := AssetParam.manager(ScadaCert.asset_id()),
            # Assert(manager.hasValue()),
            # Assert(
            #     manager.value() == config.Public().gnf_admin_addr,
            #     comment=ScadaCertErrors.WrongManager,
            # ),
            self.scada_cert_idx.set(ScadaCert.asset_id()),
            self.ta_alias.set(ta_alias.value()),
            output.set(ta_alias.value()),
        )

    @external
    def bootstrap2(
        self,
        atn_seed: abi.PaymentTransaction,
        TaTradingRights: abi.Asset,
        *,
        output: abi.String,
    ):
        """
        "Second half of the bootstrap, done by Atn calling this method. Method checks that
           - TaTradingRights are valid, and match the global
           ta_alias (which must already exist)
           - Atn payment meets the MinimumBalance requirements of the DispatchContract
           - payment sender owns the TaTradingRights, thus proving their
           identity as AtomicTNode

            https://gridworks.readthedocs.io/en/latest/ta-trading-rights.html
            If these checks pass, then the global atn_addr is set.

        Args:
            atn_seed: Initial Payment transaction to the app from Atn. This pays for boxes.
            TaTradingRights: The asset owned by the Atn Addr validating is right to trade
            energy on behalf of the TerminalAsset

        Returns:
            The TerminalAssetAlias
        """
        return Seq(
            Assert(
                atn_seed.get().receiver() == self.address,
                comment=DispatchContractErrors.ReceiverNotAppAddr,
            ),
            cert_type := AssetParam.unitName(TaTradingRights.asset_id()),
            Assert(cert_type.hasValue()),
            Assert(
                cert_type.value() == Bytes("TATRADE"),
                comment=AtnCertErrors.WrongUnitName,
            ),
            ta_alias := AssetParam.name(TaTradingRights.asset_id()),
            Assert(ta_alias.hasValue()),
            Assert(self.ta_alias == ta_alias.value()),
            atn_cert_balance := AssetHolding.balance(
                atn_seed.get().sender(), TaTradingRights.asset_id()
            ),
            Assert(
                atn_cert_balance.hasValue(),
                comment=DispatchContractErrors.AtnDoesNotOwnTradingRights,
            ),
            Assert(
                atn_cert_balance.value() == Int(1),
                comment=DispatchContractErrors.AtnDoesNotOwnTradingRights,
            ),
            self.ta_trading_rights_idx.set(TaTradingRights.asset_id()),
            output.set(self.ta_alias),
        )

    @external
    def get_ta_alias(self, *, output: abi.String):
        """
        Returns the GNodeAlias of the TerminalAsset for the Dispatch Contract.
        https://gridworks.readthedocs.io/en/latest/g-node-alias.html
        https://gridworks.readthedocs.io/en/latest/terminal-asset.html

        Returns:
            TerminalAsset GNodeAlias
        """
        return Seq(
            output.set(self.ta_alias.get()),
        )

    @create
    def create(self):
        return self.initialize_application_state()

    @opt_in
    def opt_in(self):
        """Address opting in must be either scada_addr or atn_addr"""
        # Todo: figure out how to get the address opting in
        return self.initialize_account_state()

    @external(authorize=Authorize.opted_in())
    def heartbeat_algo_audit(
        self,
        signed_proof: abi.PaymentTransaction,
        heartbeat: HeartbeatB,
        *,
        output: abi.String,
    ):
        """
        What: Algo payload with report of last HeartbeatB sent to partner via Rabbit, to be sent to
        DispatchContract on Algo blockchain

        Why: After validating the identity of the sender by inspecting the signed_proof, the
        DispatchContract shortens up the heartbeat info into the more compact form of
        HeartbeatStorageData, and stores in a box. Two hours of the most recent minutely heartbeats are
        stored, using one box for recent Atn heartbeats and one box for recent Scada heartbeats.

        See https://gridworks-atn.readthedocs.io/en/latest/apis/types.html#heartbeatalgoaudit

        Args:
            signed_proof: Tiny signed payment to DispatchContract to prove identity
            heartbeat: Heartbeat sender last sent to its partner
            See https://gridworks-atn.readthedocs.io/en/latest/apis/types.html#heartbeatb
        """
        return Seq(
            output.set(self.ta_alias.get()),
        )

    @close_out
    def close_out(self):
        return Approve()

    @delete(authorize=Authorize.only(governor))
    def delete(self):
        return Approve()


if __name__ == "__main__":
    DispatchContract().dump("dispatch_contract_artifacts")



""" HeatPumpWithBoostStore DNode Definition
"""
from typing import Optional

from gwatn.data_classes.d_node import DNode


SIG_FIGS_FOR_OUTPUT = 6


class Node_BrickStorageHeater(DNode):
    def __init__(
        self,
        ts_idx: int,
        store_idx: int,
        store_enthalpy_kwh: Optional[int] = None,
        store_avg_brick_temp_c: Optional[float] = None,
    ):
        DNode.__init__(
            self,
            ts_idx=ts_idx,
            store_idx=store_idx,
        )
        self.store_enthalpy_kwh = store_enthalpy_kwh
        self.store_avg_brick_temp_c = store_avg_brick_temp_c

    def __repr__(
        self,
    ) -> str:
        rep = f"DNode => TimeSliceIdx: {self.ts_idx}, StoreIdx: {self.store_idx}"
        if self.path_cost:
            rep += f", Path cost: ${round(self.path_cost, 3)}"
        return rep



"""Type market.type.gt, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.data_classes import MarketType
from gwatn.enums import MarketPriceUnit
from gwatn.enums import MarketQuantityUnit
from gwatn.enums import MarketTypeName
from gwatn.enums import RecognizedCurrencyUnit


class MarketPriceUnit000SchemaEnum:
    enum_name: str = "market.price.unit.000"
    symbols: List[str] = [
        "00000000",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketPriceUnit000(StrEnum):
    USDPerMWh = auto()

    @classmethod
    def default(cls) -> "MarketPriceUnit000":
        return cls.USDPerMWh

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketPriceUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketPriceUnit:
        if not MarketPriceUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketPriceUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketPriceUnit, MarketPriceUnit.default())

    @classmethod
    def local_to_type(cls, market_price_unit: MarketPriceUnit) -> str:
        if not isinstance(market_price_unit, MarketPriceUnit):
            raise SchemaError(f"{market_price_unit} must be of type {MarketPriceUnit}")
        versioned_enum = as_enum(
            market_price_unit, MarketPriceUnit000, MarketPriceUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketPriceUnit000] = {
        "00000000": MarketPriceUnit000.USDPerMWh,
    }

    versioned_enum_to_type_dict: Dict[MarketPriceUnit000, str] = {
        MarketPriceUnit000.USDPerMWh: "00000000",
    }


class RecognizedCurrencyUnit000SchemaEnum:
    enum_name: str = "recognized.currency.unit.000"
    symbols: List[str] = [
        "00000000",
        "e57c5143",
        "f7b38fc5",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class RecognizedCurrencyUnit000(StrEnum):
    UNKNOWN = auto()
    USD = auto()
    GBP = auto()

    @classmethod
    def default(cls) -> "RecognizedCurrencyUnit000":
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class RecognizedCurrencyUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> RecognizedCurrencyUnit:
        if not RecognizedCurrencyUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to RecognizedCurrencyUnit000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )

    @classmethod
    def local_to_type(cls, recognized_currency_unit: RecognizedCurrencyUnit) -> str:
        if not isinstance(recognized_currency_unit, RecognizedCurrencyUnit):
            raise SchemaError(
                f"{recognized_currency_unit} must be of type {RecognizedCurrencyUnit}"
            )
        versioned_enum = as_enum(
            recognized_currency_unit,
            RecognizedCurrencyUnit000,
            RecognizedCurrencyUnit000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, RecognizedCurrencyUnit000] = {
        "00000000": RecognizedCurrencyUnit000.UNKNOWN,
        "e57c5143": RecognizedCurrencyUnit000.USD,
        "f7b38fc5": RecognizedCurrencyUnit000.GBP,
    }

    versioned_enum_to_type_dict: Dict[RecognizedCurrencyUnit000, str] = {
        RecognizedCurrencyUnit000.UNKNOWN: "00000000",
        RecognizedCurrencyUnit000.USD: "e57c5143",
        RecognizedCurrencyUnit000.GBP: "f7b38fc5",
    }


class MarketTypeName000SchemaEnum:
    enum_name: str = "market.type.name.000"
    symbols: List[str] = [
        "00000000",
        "d20b81e4",
        "b36cbfb4",
        "94a3fe9b",
        "5f335bdb",
        "01a84101",
        "e997ccfb",
        "618f9c0a",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketTypeName000(StrEnum):
    unknown = auto()
    rt5gate5 = auto()
    rt60gate5 = auto()
    da60 = auto()
    rt60gate30 = auto()
    rt15gate5 = auto()
    rt30gate5 = auto()
    rt60gate30b = auto()

    @classmethod
    def default(cls) -> "MarketTypeName000":
        return cls.unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketTypeNameMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketTypeName:
        if not MarketTypeName000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketTypeName000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketTypeName, MarketTypeName.default())

    @classmethod
    def local_to_type(cls, market_type_name: MarketTypeName) -> str:
        if not isinstance(market_type_name, MarketTypeName):
            raise SchemaError(f"{market_type_name} must be of type {MarketTypeName}")
        versioned_enum = as_enum(
            market_type_name, MarketTypeName000, MarketTypeName000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketTypeName000] = {
        "00000000": MarketTypeName000.unknown,
        "d20b81e4": MarketTypeName000.rt5gate5,
        "b36cbfb4": MarketTypeName000.rt60gate5,
        "94a3fe9b": MarketTypeName000.da60,
        "5f335bdb": MarketTypeName000.rt60gate30,
        "01a84101": MarketTypeName000.rt15gate5,
        "e997ccfb": MarketTypeName000.rt30gate5,
        "618f9c0a": MarketTypeName000.rt60gate30b,
    }

    versioned_enum_to_type_dict: Dict[MarketTypeName000, str] = {
        MarketTypeName000.unknown: "00000000",
        MarketTypeName000.rt5gate5: "d20b81e4",
        MarketTypeName000.rt60gate5: "b36cbfb4",
        MarketTypeName000.da60: "94a3fe9b",
        MarketTypeName000.rt60gate30: "5f335bdb",
        MarketTypeName000.rt15gate5: "01a84101",
        MarketTypeName000.rt30gate5: "e997ccfb",
        MarketTypeName000.rt60gate30b: "618f9c0a",
    }


class MarketQuantityUnit000SchemaEnum:
    enum_name: str = "market.quantity.unit.000"
    symbols: List[str] = [
        "00000000",
        "c272f3b3",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketQuantityUnit000(StrEnum):
    AvgMW = auto()
    AvgkW = auto()

    @classmethod
    def default(cls) -> "MarketQuantityUnit000":
        return cls.AvgMW

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketQuantityUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketQuantityUnit:
        if not MarketQuantityUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketQuantityUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketQuantityUnit, MarketQuantityUnit.default())

    @classmethod
    def local_to_type(cls, market_quantity_unit: MarketQuantityUnit) -> str:
        if not isinstance(market_quantity_unit, MarketQuantityUnit):
            raise SchemaError(
                f"{market_quantity_unit} must be of type {MarketQuantityUnit}"
            )
        versioned_enum = as_enum(
            market_quantity_unit, MarketQuantityUnit000, MarketQuantityUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketQuantityUnit000] = {
        "00000000": MarketQuantityUnit000.AvgMW,
        "c272f3b3": MarketQuantityUnit000.AvgkW,
    }

    versioned_enum_to_type_dict: Dict[MarketQuantityUnit000, str] = {
        MarketQuantityUnit000.AvgMW: "00000000",
        MarketQuantityUnit000.AvgkW: "c272f3b3",
    }


class MarketTypeGt(BaseModel):
    """Used by MarketMakers to simultaneously run several different types of Markets.

        A [MarketMaker](https://gridworks.readthedocs.io/en/latest/market-maker.html) GNode can run several types of Markets. For example, it can run an
    hourly real-time market and also an ancillary services market for Regulation. This is captured
    by the concept of MarketType.
        [More info](https://gridworks.readthedocs.io/en/latest/market-type.html).
    """

    Name: MarketTypeName = Field(
        title="Name of the MarketType",
    )
    DurationMinutes: int = Field(
        title="Duration of MarketSlots, in minutes",
    )
    GateClosingSeconds: int = Field(
        title="Seconds before the start of a MarketSlot after which bids are not accepted",
    )
    PriceUnit: MarketPriceUnit = Field(
        title="Price Unit for market (e.g. USD Per MWh)",
    )
    QuantityUnit: MarketQuantityUnit = Field(
        title="Quantity Unit for market (e.g. AvgMW)",
    )
    CurrencyUnit: RecognizedCurrencyUnit = Field(
        title="Currency Unit for market (e.g. USD)",
    )
    PriceMax: int = Field(
        title="PMax, required for defining bids",
    )
    TypeName: Literal["market.type.gt"] = "market.type.gt"
    Version: str = "000"

    @validator("Name")
    def _check_name(cls, v: MarketTypeName) -> MarketTypeName:
        return as_enum(v, MarketTypeName, MarketTypeName.unknown)

    @validator("PriceUnit")
    def _check_price_unit(cls, v: MarketPriceUnit) -> MarketPriceUnit:
        return as_enum(v, MarketPriceUnit, MarketPriceUnit.USDPerMWh)

    @validator("QuantityUnit")
    def _check_quantity_unit(cls, v: MarketQuantityUnit) -> MarketQuantityUnit:
        return as_enum(v, MarketQuantityUnit, MarketQuantityUnit.AvgMW)

    @validator("CurrencyUnit")
    def _check_currency_unit(cls, v: RecognizedCurrencyUnit) -> RecognizedCurrencyUnit:
        return as_enum(v, RecognizedCurrencyUnit, RecognizedCurrencyUnit.UNKNOWN)

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Name"]
        Name = as_enum(self.Name, MarketTypeName, MarketTypeName.default())
        d["NameGtEnumSymbol"] = MarketTypeNameMap.local_to_type(Name)
        del d["PriceUnit"]
        PriceUnit = as_enum(self.PriceUnit, MarketPriceUnit, MarketPriceUnit.default())
        d["PriceUnitGtEnumSymbol"] = MarketPriceUnitMap.local_to_type(PriceUnit)
        del d["QuantityUnit"]
        QuantityUnit = as_enum(
            self.QuantityUnit, MarketQuantityUnit, MarketQuantityUnit.default()
        )
        d["QuantityUnitGtEnumSymbol"] = MarketQuantityUnitMap.local_to_type(
            QuantityUnit
        )
        del d["CurrencyUnit"]
        CurrencyUnit = as_enum(
            self.CurrencyUnit, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )
        d["CurrencyUnitGtEnumSymbol"] = RecognizedCurrencyUnitMap.local_to_type(
            CurrencyUnit
        )
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class MarketTypeGt_Maker:
    type_name = "market.type.gt"
    version = "000"

    def __init__(
        self,
        name: MarketTypeName,
        duration_minutes: int,
        gate_closing_seconds: int,
        price_unit: MarketPriceUnit,
        quantity_unit: MarketQuantityUnit,
        currency_unit: RecognizedCurrencyUnit,
        price_max: int,
    ):
        self.tuple = MarketTypeGt(
            Name=name,
            DurationMinutes=duration_minutes,
            GateClosingSeconds=gate_closing_seconds,
            PriceUnit=price_unit,
            QuantityUnit=quantity_unit,
            CurrencyUnit=currency_unit,
            PriceMax=price_max,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: MarketTypeGt) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> MarketTypeGt:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> MarketTypeGt:
        d2 = dict(d)
        if "NameGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing NameGtEnumSymbol")
        if d2["NameGtEnumSymbol"] in MarketTypeName000SchemaEnum.symbols:
            d2["Name"] = MarketTypeNameMap.type_to_local(d2["NameGtEnumSymbol"])
        else:
            d2["Name"] = MarketTypeName.default()
        if "DurationMinutes" not in d2.keys():
            raise SchemaError(f"dict {d2} missing DurationMinutes")
        if "GateClosingSeconds" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GateClosingSeconds")
        if "PriceUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceUnitGtEnumSymbol")
        if d2["PriceUnitGtEnumSymbol"] in MarketPriceUnit000SchemaEnum.symbols:
            d2["PriceUnit"] = MarketPriceUnitMap.type_to_local(
                d2["PriceUnitGtEnumSymbol"]
            )
        else:
            d2["PriceUnit"] = MarketPriceUnit.default()
        if "QuantityUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing QuantityUnitGtEnumSymbol")
        if d2["QuantityUnitGtEnumSymbol"] in MarketQuantityUnit000SchemaEnum.symbols:
            d2["QuantityUnit"] = MarketQuantityUnitMap.type_to_local(
                d2["QuantityUnitGtEnumSymbol"]
            )
        else:
            d2["QuantityUnit"] = MarketQuantityUnit.default()
        if "CurrencyUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing CurrencyUnitGtEnumSymbol")
        if (
            d2["CurrencyUnitGtEnumSymbol"]
            in RecognizedCurrencyUnit000SchemaEnum.symbols
        ):
            d2["CurrencyUnit"] = RecognizedCurrencyUnitMap.type_to_local(
                d2["CurrencyUnitGtEnumSymbol"]
            )
        else:
            d2["CurrencyUnit"] = RecognizedCurrencyUnit.default()
        if "PriceMax" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceMax")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return MarketTypeGt(
            Name=d2["Name"],
            DurationMinutes=d2["DurationMinutes"],
            GateClosingSeconds=d2["GateClosingSeconds"],
            PriceUnit=d2["PriceUnit"],
            QuantityUnit=d2["QuantityUnit"],
            CurrencyUnit=d2["CurrencyUnit"],
            PriceMax=d2["PriceMax"],
            TypeName=d2["TypeName"],
            Version="000",
        )

    @classmethod
    def tuple_to_dc(cls, t: MarketTypeGt) -> MarketType:
        if t.Name in MarketType.by_id.keys():
            dc = MarketType.by_id[t.Name]
        else:
            dc = MarketType(
                name=t.Name,
                duration_minutes=t.DurationMinutes,
                gate_closing_seconds=t.GateClosingSeconds,
                price_unit=t.PriceUnit,
                quantity_unit=t.QuantityUnit,
                currency_unit=t.CurrencyUnit,
                price_max=t.PriceMax,
            )

        return dc

    @classmethod
    def dc_to_tuple(cls, dc: MarketType) -> MarketTypeGt:
        t = MarketTypeGt_Maker(
            name=dc.name,
            duration_minutes=dc.duration_minutes,
            gate_closing_seconds=dc.gate_closing_seconds,
            price_unit=dc.price_unit,
            quantity_unit=dc.quantity_unit,
            currency_unit=dc.currency_unit,
            price_max=dc.price_max,
        ).tuple
        return t

    @classmethod
    def type_to_dc(cls, t: str) -> MarketType:
        return cls.tuple_to_dc(cls.type_to_tuple(t))

    @classmethod
    def dc_to_type(cls, dc: MarketType) -> str:
        return cls.dc_to_tuple(dc).as_type()

    @classmethod
    def dict_to_dc(cls, d: dict[Any, str]) -> MarketType:
        return cls.tuple_to_dc(cls.dict_to_tuple(d))



""" List of all the types used"""
from typing import Dict
from typing import List
from typing import no_type_check

from gwatn.types import AcceptedBid_Maker
from gwatn.types import AtnBid_Maker
from gwatn.types import AtnParams_Maker
from gwatn.types import AtnParamsBrickstorageheater_Maker
from gwatn.types import AtnParamsReport_Maker
from gwatn.types import BaseGNodeGt_Maker
from gwatn.types import BasegnodeScadaCreate_Maker
from gwatn.types import ComponentAttributeClassGt_Maker
from gwatn.types import ComponentGt_Maker
from gwatn.types import DataChannel_Maker
from gwatn.types import DiscoverycertAlgoCreate_Maker
from gwatn.types import DispatchContractConfirmed_Maker
from gwatn.types import EgaugeIo_Maker
from gwatn.types import EgaugeRegisterConfig_Maker
from gwatn.types import ElectricMeterCacGt_Maker
from gwatn.types import ElectricMeterComponentGt_Maker
from gwatn.types import FloParams_Maker
from gwatn.types import FloParamsBrickstorageheater_Maker
from gwatn.types import FloParamsReport_Maker
from gwatn.types import GNodeGt_Maker
from gwatn.types import GNodeInstanceGt_Maker
from gwatn.types import GtDispatchBoolean_Maker
from gwatn.types import GtDispatchBooleanLocal_Maker
from gwatn.types import GtDriverBooleanactuatorCmd_Maker
from gwatn.types import GtShBooleanactuatorCmdStatus_Maker
from gwatn.types import GtShCliAtnCmd_Maker
from gwatn.types import GtShMultipurposeTelemetryStatus_Maker
from gwatn.types import GtShSimpleTelemetryStatus_Maker
from gwatn.types import GtShStatus_Maker
from gwatn.types import GtShTelemetryFromMultipurposeSensor_Maker
from gwatn.types import GtTelemetry_Maker
from gwatn.types import GwCertId_Maker
from gwatn.types import HeartbeatA_Maker
from gwatn.types import HeartbeatB_Maker
from gwatn.types import InitialTadeedAlgoCreate_Maker
from gwatn.types import InitialTadeedAlgoOptin_Maker
from gwatn.types import InitialTadeedAlgoTransfer_Maker
from gwatn.types import JoinDispatchContract_Maker
from gwatn.types import LatestPrice_Maker
from gwatn.types import MarketSlot_Maker
from gwatn.types import MarketTypeGt_Maker
from gwatn.types import MultipurposeSensorCacGt_Maker
from gwatn.types import MultipurposeSensorComponentGt_Maker
from gwatn.types import NewTadeedAlgoOptin_Maker
from gwatn.types import NewTadeedSend_Maker
from gwatn.types import OldTadeedAlgoReturn_Maker
from gwatn.types import PipeFlowSensorCacGt_Maker
from gwatn.types import PipeFlowSensorComponentGt_Maker
from gwatn.types import PowerWatts_Maker
from gwatn.types import PriceQuantity_Maker
from gwatn.types import PriceQuantityUnitless_Maker
from gwatn.types import Ready_Maker
from gwatn.types import RelayCacGt_Maker
from gwatn.types import RelayComponentGt_Maker
from gwatn.types import ResistiveHeaterCacGt_Maker
from gwatn.types import ResistiveHeaterComponentGt_Maker
from gwatn.types import ScadaCertTransfer_Maker
from gwatn.types import SimplesimDriverData_Maker
from gwatn.types import SimplesimDriverDataBsh_Maker
from gwatn.types import SimplesimDriverReport_Maker
from gwatn.types import SimplesimSnapshotBrickstorageheater_Maker
from gwatn.types import SimpleTempSensorCacGt_Maker
from gwatn.types import SimpleTempSensorComponentGt_Maker
from gwatn.types import SimTimestep_Maker
from gwatn.types import SlaEnter_Maker
from gwatn.types import SnapshotSpaceheat_Maker
from gwatn.types import SpaceheatNodeGt_Maker
from gwatn.types import SuperStarter_Maker
from gwatn.types import SupervisorContainerGt_Maker
from gwatn.types import TadeedSpecsHack_Maker
from gwatn.types import TavalidatorcertAlgoCreate_Maker
from gwatn.types import TavalidatorcertAlgoTransfer_Maker
from gwatn.types import TelemetryReportingConfig_Maker
from gwatn.types import TelemetrySnapshotSpaceheat_Maker
from gwatn.types import TerminalassetCertifyHack_Maker


TypeMakerByName: Dict[str, HeartbeatA_Maker] = {}


@no_type_check
def type_makers() -> List[HeartbeatA_Maker]:
    return [
        AcceptedBid_Maker,
        AtnBid_Maker,
        AtnParams_Maker,
        AtnParamsBrickstorageheater_Maker,
        AtnParamsReport_Maker,
        BaseGNodeGt_Maker,
        BasegnodeScadaCreate_Maker,
        ComponentAttributeClassGt_Maker,
        ComponentGt_Maker,
        DataChannel_Maker,
        DiscoverycertAlgoCreate_Maker,
        DispatchContractConfirmed_Maker,
        EgaugeIo_Maker,
        EgaugeRegisterConfig_Maker,
        ElectricMeterCacGt_Maker,
        ElectricMeterComponentGt_Maker,
        FloParams_Maker,
        FloParamsBrickstorageheater_Maker,
        FloParamsReport_Maker,
        GNodeGt_Maker,
        GNodeInstanceGt_Maker,
        GtDispatchBoolean_Maker,
        GtDispatchBooleanLocal_Maker,
        GtDriverBooleanactuatorCmd_Maker,
        GtShBooleanactuatorCmdStatus_Maker,
        GtShCliAtnCmd_Maker,
        GtShMultipurposeTelemetryStatus_Maker,
        GtShSimpleTelemetryStatus_Maker,
        GtShStatus_Maker,
        GtShTelemetryFromMultipurposeSensor_Maker,
        GtTelemetry_Maker,
        GwCertId_Maker,
        HeartbeatA_Maker,
        HeartbeatB_Maker,
        InitialTadeedAlgoCreate_Maker,
        InitialTadeedAlgoOptin_Maker,
        InitialTadeedAlgoTransfer_Maker,
        JoinDispatchContract_Maker,
        LatestPrice_Maker,
        MarketSlot_Maker,
        MarketTypeGt_Maker,
        MultipurposeSensorCacGt_Maker,
        MultipurposeSensorComponentGt_Maker,
        NewTadeedAlgoOptin_Maker,
        NewTadeedSend_Maker,
        OldTadeedAlgoReturn_Maker,
        PipeFlowSensorCacGt_Maker,
        PipeFlowSensorComponentGt_Maker,
        PowerWatts_Maker,
        PriceQuantity_Maker,
        PriceQuantityUnitless_Maker,
        Ready_Maker,
        RelayCacGt_Maker,
        RelayComponentGt_Maker,
        ResistiveHeaterCacGt_Maker,
        ResistiveHeaterComponentGt_Maker,
        ScadaCertTransfer_Maker,
        SimTimestep_Maker,
        SimpleTempSensorCacGt_Maker,
        SimpleTempSensorComponentGt_Maker,
        SimplesimDriverData_Maker,
        SimplesimDriverDataBsh_Maker,
        SimplesimDriverReport_Maker,
        SimplesimSnapshotBrickstorageheater_Maker,
        SlaEnter_Maker,
        SnapshotSpaceheat_Maker,
        SpaceheatNodeGt_Maker,
        SuperStarter_Maker,
        SupervisorContainerGt_Maker,
        TadeedSpecsHack_Maker,
        TavalidatorcertAlgoCreate_Maker,
        TavalidatorcertAlgoTransfer_Maker,
        TelemetryReportingConfig_Maker,
        TelemetrySnapshotSpaceheat_Maker,
        TerminalassetCertifyHack_Maker,
    ]


for maker in type_makers():
    TypeMakerByName[maker.type_name] = maker


def version_by_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are TypeNames, values are versions
    """

    v: Dict[str, str] = {
        "accepted.bid": "000",
        "atn.bid": "001",
        "atn.params": "000",
        "atn.params.brickstorageheater": "000",
        "atn.params.report": "000",
        "base.g.node.gt": "002",
        "basegnode.scada.create": "000",
        "component.attribute.class.gt": "000",
        "component.gt": "000",
        "data.channel": "000",
        "discoverycert.algo.create": "000",
        "dispatch.contract.confirmed": "000",
        "egauge.io": "000",
        "egauge.register.config": "000",
        "electric.meter.cac.gt": "000",
        "electric.meter.component.gt": "000",
        "flo.params": "000",
        "flo.params.brickstorageheater": "000",
        "flo.params.report": "000",
        "g.node.gt": "002",
        "g.node.instance.gt": "000",
        "gt.dispatch.boolean": "110",
        "gt.dispatch.boolean.local": "110",
        "gt.driver.booleanactuator.cmd": "100",
        "gt.sh.booleanactuator.cmd.status": "100",
        "gt.sh.cli.atn.cmd": "110",
        "gt.sh.multipurpose.telemetry.status": "100",
        "gt.sh.simple.telemetry.status": "100",
        "gt.sh.status": "110",
        "gt.sh.telemetry.from.multipurpose.sensor": "100",
        "gt.telemetry": "110",
        "gw.cert.id": "000",
        "heartbeat.a": "100",
        "heartbeat.b": "001",
        "initial.tadeed.algo.create": "000",
        "initial.tadeed.algo.optin": "002",
        "initial.tadeed.algo.transfer": "000",
        "join.dispatch.contract": "000",
        "latest.price": "000",
        "market.slot": "000",
        "market.type.gt": "000",
        "multipurpose.sensor.cac.gt": "000",
        "multipurpose.sensor.component.gt": "000",
        "new.tadeed.algo.optin": "000",
        "new.tadeed.send": "000",
        "old.tadeed.algo.return": "000",
        "pipe.flow.sensor.cac.gt": "000",
        "pipe.flow.sensor.component.gt": "000",
        "power.watts": "000",
        "price.quantity": "000",
        "price.quantity.unitless": "000",
        "ready": "001",
        "relay.cac.gt": "000",
        "relay.component.gt": "000",
        "resistive.heater.cac.gt": "000",
        "resistive.heater.component.gt": "000",
        "scada.cert.transfer": "000",
        "sim.timestep": "000",
        "simple.temp.sensor.cac.gt": "000",
        "simple.temp.sensor.component.gt": "000",
        "simplesim.driver.data": "000",
        "simplesim.driver.data.bsh": "000",
        "simplesim.driver.report": "000",
        "simplesim.snapshot.brickstorageheater": "000",
        "sla.enter": "000",
        "snapshot.spaceheat": "000",
        "spaceheat.node.gt": "100",
        "super.starter": "000",
        "supervisor.container.gt": "000",
        "tadeed.specs.hack": "000",
        "tavalidatorcert.algo.create": "000",
        "tavalidatorcert.algo.transfer": "000",
        "telemetry.reporting.config": "000",
        "telemetry.snapshot.spaceheat": "000",
        "terminalasset.certify.hack": "000",
    }

    return v


def status_by_versioned_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are versioned TypeNames, values are type status
    """

    v: Dict[str, str] = {
        "accepted.bid.000": "Pending",
        "atn.bid.001": "Pending",
        "atn.params.000": "Active",
        "atn.params.brickstorageheater.000": "Pending",
        "atn.params.report.000": "Active",
        "base.g.node.gt.002": "Active",
        "basegnode.scada.create.000": "Active",
        "component.attribute.class.gt.000": "Active",
        "component.gt.000": "Active",
        "data.channel.000": "Active",
        "discoverycert.algo.create.000": "Active",
        "dispatch.contract.confirmed.000": "Active",
        "egauge.io.000": "Active",
        "egauge.register.config.000": "Active",
        "electric.meter.cac.gt.000": "Active",
        "electric.meter.component.gt.000": "Active",
        "flo.params.000": "Active",
        "flo.params.brickstorageheater.000": "Active",
        "flo.params.report.000": "Pending",
        "g.node.gt.002": "Active",
        "g.node.instance.gt.000": "Active",
        "gt.dispatch.boolean.110": "Active",
        "gt.dispatch.boolean.local.110": "Active",
        "gt.driver.booleanactuator.cmd.100": "Active",
        "gt.sh.booleanactuator.cmd.status.100": "Active",
        "gt.sh.cli.atn.cmd.110": "Active",
        "gt.sh.multipurpose.telemetry.status.100": "Active",
        "gt.sh.simple.telemetry.status.100": "Active",
        "gt.sh.status.110": "Active",
        "gt.sh.telemetry.from.multipurpose.sensor.100": "Active",
        "gt.telemetry.110": "Active",
        "gw.cert.id.000": "Active",
        "heartbeat.a.100": "Active",
        "heartbeat.b.001": "Active",
        "initial.tadeed.algo.create.000": "Active",
        "initial.tadeed.algo.optin.002": "Active",
        "initial.tadeed.algo.transfer.000": "Active",
        "join.dispatch.contract.000": "Active",
        "latest.price.000": "Pending",
        "market.slot.000": "Pending",
        "market.type.gt.000": "Pending",
        "multipurpose.sensor.cac.gt.000": "Active",
        "multipurpose.sensor.component.gt.000": "Active",
        "new.tadeed.algo.optin.000": "Active",
        "new.tadeed.send.000": "Active",
        "old.tadeed.algo.return.000": "Active",
        "pipe.flow.sensor.cac.gt.000": "Active",
        "pipe.flow.sensor.component.gt.000": "Active",
        "power.watts.000": "Active",
        "price.quantity.000": "Pending",
        "price.quantity.unitless.000": "Pending",
        "ready.001": "Active",
        "relay.cac.gt.000": "Active",
        "relay.component.gt.000": "Active",
        "resistive.heater.cac.gt.000": "Active",
        "resistive.heater.component.gt.000": "Active",
        "scada.cert.transfer.000": "Pending",
        "sim.timestep.000": "Active",
        "simple.temp.sensor.cac.gt.000": "Active",
        "simple.temp.sensor.component.gt.000": "Active",
        "simplesim.driver.data.000": "Pending",
        "simplesim.driver.data.bsh.000": "Pending",
        "simplesim.driver.report.000": "Pending",
        "simplesim.snapshot.brickstorageheater.000": "Pending",
        "sla.enter.000": "Pending",
        "snapshot.spaceheat.000": "Active",
        "spaceheat.node.gt.100": "Active",
        "super.starter.000": "Active",
        "supervisor.container.gt.000": "Active",
        "tadeed.specs.hack.000": "Pending",
        "tavalidatorcert.algo.create.000": "Active",
        "tavalidatorcert.algo.transfer.000": "Active",
        "telemetry.reporting.config.000": "Active",
        "telemetry.snapshot.spaceheat.000": "Active",
        "terminalasset.certify.hack.000": "Pending",
    }

    return v
